from .. import _iterate
from ..attach import attach, detach
from ..duration import Duration, NonreducedFraction
from ..enums import Left, Right
from ..indicators.StartBeam import StartBeam
from ..indicators.StartSlur import StartSlur
from ..indicators.StopBeam import StopBeam
from ..indicators.StopSlur import StopSlur
from ..indicators.Tie import Tie
from ..indicators.TimeSignature import TimeSignature
from ..pitch.pitchclasses import NamedPitchClass
from ..pitch.pitches import NamedPitch
from ..score import Chord, Container, Note, Rest, Tuplet
from ..select import Selection
from .base import Parser


class ReducedLyParser(Parser):
    r"""
    Parses the "reduced-ly" syntax, a modified subset of LilyPond syntax.

    ..  container:: example

        >>> from abjad.parsers.reduced import ReducedLyParser
        >>> parser = ReducedLyParser()

        Understands LilyPond-like representation of notes, chords and rests:

        >>> string = "c'4 r8. <b d' fs'>16"
        >>> container = parser(string)
        >>> abjad.show(container) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(container)
            >>> print(string)
            {
                c'4
                r8.
                <b d' fs'>16
            }

        Also parses bare duration as notes on middle-C, and negative bare
        durations as rests:

        >>> string = '4 -8 16. -32'
        >>> container = parser(string)
        >>> abjad.show(container) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(container)
            >>> print(string)
            {
                c'4
                r8
                c'16.
                r32
            }

    ..  container:: example

        Note that the leaf syntax is greedy, and therefore duration specifiers
        following pitch specifiers will be treated as part of the same
        expression. The following produces 2 leaves, rather than 3:

        >>> string = "4 d' 4"
        >>> container = parser(string)
        >>> abjad.show(container) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(container)
            >>> print(string)
            {
                c'4
                d'4
            }

        Understands LilyPond-like default durations:

        >>> string = "c'4 d' e' f'"
        >>> container = parser(string)
        >>> abjad.show(container) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(container)
            >>> print(string)
            {
                c'4
                d'4
                e'4
                f'4
            }

    Also understands various types of container specifications.

    ..  container:: example

        Can create arbitrarily nested tuplets:

        >>> string = "2/3 { 4 4 3/5 { 8 8 8 } }"
        >>> tuplet = parser(string)
        >>> abjad.show(tuplet) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(tuplet)
            >>> print(string)
            \tweak edge-height #'(0.7 . 0)
            \times 2/3 {
                c'4
                c'4
                \tweak text #tuplet-number::calc-fraction-text
                \tweak edge-height #'(0.7 . 0)
                \times 3/5 {
                    c'8
                    c'8
                    c'8
                }
            }

    ..  container:: example

        Can create measures too:

        >>> string = '| 4/4 4 4 4 4 || 3/8 8 8 8 |'
        >>> container = parser(string)
        >>> staff = abjad.Staff([container])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                {
                    {
                        \time 4/4
                        c'4
                        c'4
                        c'4
                        c'4
                    }
                    {
                        \time 3/8
                        c'8
                        c'8
                        c'8
                    }
                }
            }

    ..  container:: example

        Finally, understands ties, slurs and beams:

        >>> string = 'c16 [ ( d ~ d ) f ]'
        >>> container = parser(string)
        >>> abjad.show(container) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(container)
            >>> print(string)
            {
                c16
                [
                (
                d16
                ~
                d16
                )
                f16
                ]
            }

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_default_duration", "_toplevel_component_count")

    ### INITIALIZER ###

    def __init__(self, debug=False):
        self._default_duration = Duration((1, 4))
        self._toplevel_component_count = None
        Parser.__init__(self, debug=debug)

    ### LEX SETUP ###

    tokens = (
        "APOSTROPHE",
        "BRACE_L",
        "BRACE_R",
        "BRACKET_L",
        "BRACKET_R",
        "CARAT_L",
        "CARAT_R",
        "COMMA",
        "DOT",
        "FRACTION",
        "INTEGER_N",
        "INTEGER_P",
        "PAREN_L",
        "PAREN_R",
        "PIPE",
        "PITCHNAME",
        "RESTNAME",
        "TILDE",
    )

    t_APOSTROPHE = "'"
    t_BRACE_L = "{"
    t_BRACE_R = "}"
    t_BRACKET_L = r"\["
    t_BRACKET_R = r"\]"
    t_CARAT_L = r"\<"
    t_CARAT_R = r"\>"
    t_COMMA = ","
    t_DOT = r"\."
    t_PAREN_L = r"\("
    t_PAREN_R = r"\)"
    t_PIPE = r"\|"
    t_RESTNAME = "r"
    t_TILDE = "~"

    t_ignore = " \t\r"

    ### YACC SETUP ###

    start = "start"

    ### LEX METHODS ###

    def t_FRACTION(self, t):
        r"([1-9]\d*/[1-9]\d*)"
        parts = t.value.split("/")
        t.value = NonreducedFraction(int(parts[0]), int(parts[1]))
        return t

    def t_INTEGER_N(self, t):
        r"(-[1-9]\d*)"
        t.value = int(t.value)
        return t

    def t_INTEGER_P(self, t):
        r"([1-9]\d*)"
        t.value = int(t.value)
        return t

    def t_PITCHNAME(self, t):
        r"[a-g](ff|ss|f|s|tqf|tqs|qs|qf)?"
        t.value = NamedPitchClass(t.value)
        return t

    def t_error(self, t):
        print(("Illegal character '%s'" % t.value[0]))
        t.lexer.skip(1)

    def t_newline(self, t):
        r"\n+"
        t.lexer.lineno += t.value.count("\n")

    ### YACC METHODS ###

    def p_apostrophes__APOSTROPHE(self, p):
        """
        apostrophes : APOSTROPHE
        """
        p[0] = 1

    def p_apostrophes__apostrophes__APOSTROPHE(self, p):
        """
        apostrophes : apostrophes APOSTROPHE
        """
        p[0] = p[1] + 1

    def p_beam__BRACKET_L(self, p):
        """
        beam : BRACKET_L
        """
        p[0] = (StartBeam, Left)

    def p_beam__BRACKET_R(self, p):
        """
        beam : BRACKET_R
        """
        p[0] = (StopBeam, Right)

    def p_chord_body__chord_pitches(self, p):
        """
        chord_body : chord_pitches
        """
        p[0] = Chord(p[1], self._default_duration)

    def p_chord_body__chord_pitches__positive_leaf_duration(self, p):
        """
        chord_body : chord_pitches positive_leaf_duration
        """
        p[0] = Chord(p[1], p[2])

    def p_chord_pitches__CARAT_L__pitches__CARAT_R(self, p):
        """
        chord_pitches : CARAT_L pitches CARAT_R
        """
        p[0] = p[2]

    def p_commas__COMMA(self, p):
        """
        commas : COMMA
        """
        p[0] = 1

    def p_commas__commas__commas(self, p):
        """
        commas : commas COMMA
        """
        p[0] = p[1] + 1

    def p_component__container(self, p):
        """
        component : container
        """
        p[0] = p[1]

    def p_component__fixed_duration_container(self, p):
        """
        component : fixed_duration_container
        """
        p[0] = p[1]

    def p_component__leaf(self, p):
        """
        component : leaf
        """
        p[0] = p[1]

    def p_component__tuplet(self, p):
        """
        component : tuplet
        """
        p[0] = p[1]

    def p_component_list__EMPTY(self, p):
        """
        component_list :
        """
        p[0] = []

    def p_component_list__component_list__component(self, p):
        """
        component_list : component_list component
        """
        p[0] = p[1] + [p[2]]

    def p_container__BRACE_L__component_list__BRACE_R(self, p):
        r"""container : BRACE_L component_list BRACE_R"""
        p[0] = Container()
        for component in p[2]:
            p[0].append(component)

    def p_dots__EMPTY(self, p):
        """
        dots :
        """
        p[0] = 0

    def p_dots__dots__DOT(self, p):
        """
        dots : dots DOT
        """
        p[0] = p[1] + 1

    def p_error(self, p):
        if p:
            print(("Syntax error at '%s'" % p.value))
        else:
            print("Syntax error at EOF")

    def p_fixed_duration_container__BRACE_L__FRACTION__BRACE_R(self, p):
        """
        fixed_duration_container : BRACE_L FRACTION BRACE_R
        """
        raise Exception("fixed-duration containers no longer supported.")

    def p_leaf__leaf_body__post_events(self, p):
        """
        leaf : leaf_body post_events
        """
        p[0] = p[1]
        if p[2]:
            annotation = {"post events": p[2]}
            attach(annotation, p[0])

    def p_leaf_body__chord_body(self, p):
        """
        leaf_body : chord_body
        """
        p[0] = p[1]

    def p_leaf_body__note_body(self, p):
        """
        leaf_body : note_body
        """
        p[0] = p[1]

    def p_leaf_body__rest_body(self, p):
        """
        leaf_body : rest_body
        """
        p[0] = p[1]

    def p_measure__PIPE__FRACTION__component_list__PIPE(self, p):
        """
        measure : PIPE FRACTION component_list PIPE
        """
        measure = Container()
        for x in p[3]:
            measure.append(x)
        leaf = _iterate._get_leaf(measure, 0)
        time_signature = TimeSignature(p[2].pair)
        attach(time_signature, leaf)
        p[0] = measure

    def p_negative_leaf_duration__INTEGER_N__dots(self, p):
        """
        negative_leaf_duration : INTEGER_N dots
        """
        duration_log = p[1]
        dots = "." * p[2]
        string = f"{abs(duration_log)}{dots}"
        duration = Duration.from_lilypond_duration_string(string)
        self._default_duration = duration
        p[0] = duration

    def p_note_body__pitch(self, p):
        """
        note_body : pitch
        """
        p[0] = Note(p[1], self._default_duration)

    def p_note_body__pitch__positive_leaf_duration(self, p):
        """
        note_body : pitch positive_leaf_duration
        """
        p[0] = Note(p[1], p[2])

    def p_note_body__positive_leaf_duration(self, p):
        """
        note_body : positive_leaf_duration
        """
        p[0] = Note(0, p[1])

    def p_pitch__PITCHNAME(self, p):
        """
        pitch : PITCHNAME
        """
        p[0] = NamedPitch(str(p[1]))

    def p_pitch__PITCHNAME__apostrophes(self, p):
        """
        pitch : PITCHNAME apostrophes
        """
        p[0] = NamedPitch(str(p[1]) + "'" * p[2])

    def p_pitch__PITCHNAME__commas(self, p):
        """
        pitch : PITCHNAME commas
        """
        p[0] = NamedPitch(str(p[1]) + "," * p[2])

    def p_pitches__pitch(self, p):
        """
        pitches : pitch
        """
        p[0] = [p[1]]

    def p_pitches__pitches__pitch(self, p):
        """
        pitches : pitches pitch
        """
        p[0] = p[1] + [p[2]]

    def p_positive_leaf_duration__INTEGER_P__dots(self, p):
        """
        positive_leaf_duration : INTEGER_P dots
        """
        duration_log = p[1]
        dots = "." * p[2]
        duration = Duration.from_lilypond_duration_string(f"{abs(duration_log)}{dots}")
        self._default_duration = duration
        p[0] = duration

    def p_post_event__beam(self, p):
        """
        post_event : beam
        """
        p[0] = p[1]

    def p_post_event__slur(self, p):
        """
        post_event : slur
        """
        p[0] = p[1]

    def p_post_event__tie(self, p):
        """
        post_event : tie
        """
        p[0] = p[1]

    def p_post_events__EMPTY(self, p):
        """
        post_events :
        """
        p[0] = {}

    def p_post_events__post_events__post_event(self, p):
        """
        post_events : post_events post_event
        """
        kind, direction = p[2]
        if kind in p[1]:
            p[1][kind].append(direction)
        else:
            p[1][kind] = [direction]
        p[0] = p[1]

    def p_rest_body__RESTNAME(self, p):
        """
        rest_body : RESTNAME
        """
        p[0] = Rest(self._default_duration)

    def p_rest_body__RESTNAME__positive_leaf_duration(self, p):
        """
        rest_body : RESTNAME positive_leaf_duration
        """
        p[0] = Rest(p[2])

    def p_rest_body__negative_leaf_duration(self, p):
        """
        rest_body : negative_leaf_duration
        """
        p[0] = Rest(p[1])

    def p_slur__PAREN_L(self, p):
        """
        slur : PAREN_L
        """
        p[0] = (StartSlur, Left)

    def p_slur__PAREN_R(self, p):
        """
        slur : PAREN_R
        """
        p[0] = (StopSlur, Right)

    def p_start__EMPTY(self, p):
        """
        start :
        """
        self._toplevel_component_count = 0
        p[0] = []

    def p_start__start__component(self, p):
        """
        start : start component
        """
        self._toplevel_component_count += 1
        p[0] = p[1] + [p[2]]

    def p_start__start__measure(self, p):
        """
        start : start measure
        """
        self._toplevel_component_count += 1
        p[0] = p[1] + [p[2]]

    def p_tie__TILDE(self, p):
        """
        tie : TILDE
        """
        p[0] = (Tie, Left)

    def p_tuplet__FRACTION__container(self, p):
        """
        tuplet : FRACTION container
        """
        assert isinstance(p[2], Container)
        leaves = p[2][:]
        p[2][:] = []
        p[0] = Tuplet(p[1], leaves)

    ### PRIVATE METHODS ###

    def _attach_indicators(self, leaves):
        for leaf in leaves:
            span_events = self._get_span_events(leaf)
            for current_class, directions in span_events.items():
                if current_class in (
                    StartSlur,
                    StopSlur,
                ):
                    indicator = current_class()
                    attach(indicator, leaf)
                    continue
                if current_class in (
                    StartBeam,
                    StopBeam,
                ):
                    indicator = current_class()
                    attach(indicator, leaf)
                    continue
                if current_class is Tie:
                    indicator = current_class()
                    attach(indicator, leaf)
                    continue

    def _cleanup(self, parsed):
        container = Container()
        for x in parsed:
            container.append(x)
        parsed = container
        leaves = Selection(parsed).leaves()
        if leaves:
            self._attach_indicators(leaves)
        for leaf in leaves:
            detach(dict, leaf)
        if 1 < self._toplevel_component_count:
            return parsed
        return parsed[0]

    def _get_span_events(self, leaf):
        annotations = leaf._get_indicators(dict)
        detach(dict, leaf)
        annotations = [x for x in annotations if "post events" in x]
        if annotations:
            return annotations[0]["post events"]
        return {}

    def _setup(self):
        self._toplevel_component_count = 0
        self._default_duration = Duration((1, 4))

    ### PUBLIC PROPERTIES ###

    @property
    def debug(self):
        """
        Gets debug boolean of reduced ly parser.

        Returns true or false.
        """
        return self._debug

    @property
    def lexer_rules_object(self):
        """
        Lexer rules object of reduced ly parser.
        """
        return self

    @property
    def parser_rules_object(self):
        """
        Parser rules object of reduced ly parser.
        """
        return self


### FUNCTIONS ###


def parse_reduced_ly_syntax(string) -> Container:
    """
    Parse the reduced LilyPond rhythmic syntax:

    ..  container:: example

        >>> string = '4 -4. 8.. 5/3 { } 4'
        >>> container = abjad.parsers.reduced.parse_reduced_ly_syntax(string)
        >>> container
        <{5}>

        >>> for component in container:
        ...     component
        ...
        Note("c'4")
        Rest('r4.')
        Note("c'8..")
        Tuplet('3:5', '')
        Note("c'4")

    """
    return ReducedLyParser()(string)



import functools
import typing

from .. import typings
from ..duration import Multiplier
from ..markups import Markup
from ..storage import StorageFormatManager


@functools.total_ordering
class BowContactPoint:
    """
    Bow contact point.

    ..  container:: example

        Contact point exactly halfway from frog to tip of bow:

        >>> point = abjad.BowContactPoint((1, 2))
        >>> string = abjad.storage(point)
        >>> print(string)
        abjad.BowContactPoint(
            contact_point=abjad.Multiplier(1, 2),
            )

    ..  container:: example

        Contact point 3/5 of the way from frog to tip of bow:

        >>> point = abjad.BowContactPoint((3, 5))
        >>> string = abjad.storage(point)
        >>> print(string)
        abjad.BowContactPoint(
            contact_point=abjad.Multiplier(3, 5),
            )

    Contact points are measured from frog to tip as a fraction between ``0``
    and ``1``.
    """

    ### CLASS VARIABLES ###

    __slots__ = ("_contact_point",)

    ### INITIALIZER ###

    def __init__(self, contact_point: typings.IntegerPair = None) -> None:
        contact_point_ = None
        if contact_point is not None:
            contact_point_ = Multiplier(contact_point)
            assert 0 <= contact_point_ <= 1
        self._contact_point = contact_point_

    ### SPECIAL METHODS ###

    def __lt__(self, argument) -> bool:
        """
        Is true if ``argument`` is a bow contact point and this bow contact
        point is less than ``argument``.

        ..  container:: example

            >>> point_1 = abjad.BowContactPoint((1, 2))
            >>> point_2 = abjad.BowContactPoint((1, 2))
            >>> point_3 = abjad.BowContactPoint((2, 3))

            >>> point_1 < point_1
            False
            >>> point_1 < point_2
            False
            >>> point_1 < point_3
            True

            >>> point_2 < point_1
            False
            >>> point_2 < point_2
            False
            >>> point_2 < point_3
            True

            >>> point_3 < point_1
            False
            >>> point_3 < point_2
            False
            >>> point_3 < point_3
            False

        """
        if isinstance(argument, type(self)):
            self_contact_point = self.contact_point or 0
            argument_contact_point = argument.contact_point or 0
            return self_contact_point < argument_contact_point
        raise TypeError("unorderable types")

    def __repr__(self):
        """
        Gets interpreter representation.
        """
        return StorageFormatManager(self).get_repr_format()

    ### PUBLIC PROPERTIES ###

    @property
    def contact_point(self) -> typing.Optional[Multiplier]:
        """
        Gets contact point of bow contact point.

        ..  container:: example

            One quarter of the way from frog to point:

            >>> point = abjad.BowContactPoint((1, 4))
            >>> point.contact_point
            Multiplier(1, 4)

        ..  container:: example

            Three fifths of the way from frog to point:

            >>> point = abjad.BowContactPoint((3, 5))
            >>> point.contact_point
            Multiplier(3, 5)

        """
        return self._contact_point

    @property
    def markup(self) -> Markup:
        r"""
        Gets markup of bow contact point.

        ..  container:: example

            One quarter of the way from frog to point:

            >>> indicator = abjad.BowContactPoint((1, 4))
            >>> print(abjad.lilypond(indicator.markup))
            \markup {
                \center-align
                    \vcenter
                        \fraction
                            1
                            4
                }
            >>> abjad.show(indicator.markup) # doctest: +SKIP

        ..  container:: example

            Three fifths of the way from frog to point:

            >>> indicator = abjad.BowContactPoint((3, 5))
            >>> print(abjad.lilypond(indicator.markup))
            \markup {
                \center-align
                    \vcenter
                        \fraction
                            3
                            5
                }
            >>> abjad.show(indicator.markup) # doctest: +SKIP

        """
        if self.contact_point is None:
            contact_point = Multiplier(0, 1)
        else:
            contact_point = self.contact_point
        fraction = fr"\fraction {contact_point.numerator} {contact_point.denominator}"
        string = rf"\center-align \vcenter {fraction}"
        markup = Markup(string)
        return markup

    @property
    def tweaks(self) -> None:
        """
        Are not implemented on bow contact point.
        """
        pass



import typing

from ..storage import StorageFormatManager


class BowPressure:
    """
    Bow pressure indicator.

    ..  container:: example

        >>> bow_pressure = abjad.BowPressure('overpressure')
        >>> string = abjad.storage(bow_pressure)
        >>> print(string)
        abjad.BowPressure(
            pressure='overpressure',
            )

        >>> bow_pressure = abjad.BowPressure('underpressure')
        >>> string = abjad.storage(bow_pressure)
        >>> print(string)
        abjad.BowPressure(
            pressure='underpressure',
            )

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_pressure",)

    _parameter = "BOW_PRESSURE"

    _persistent = True

    ### INITIALIZER ###

    def __init__(self, pressure: str = None) -> None:
        self._pressure = pressure

    ### SPECIAL METHODS ###

    def __eq__(self, argument) -> bool:
        """
        Is true when all initialization values of Abjad value object equal
        the initialization values of ``argument``.
        """
        return StorageFormatManager.compare_objects(self, argument)

    def __hash__(self) -> int:
        """
        Hashes Abjad value object.
        """
        hash_values = StorageFormatManager(self).get_hash_values()
        try:
            result = hash(hash_values)
        except TypeError:
            raise TypeError(f"unhashable type: {self}")
        return result

    def __repr__(self) -> str:
        """
        Gets interpreter representation.
        """
        return StorageFormatManager(self).get_repr_format()

    ### PUBLIC PROPERTIES ###

    @property
    def parameter(self) -> str:
        """
        Returns ``'BOW_PRESSURE'``.

        ..  container:: example

            >>> abjad.BowPressure('overpressure').parameter
            'BOW_PRESSURE'

        Class constant.
        """
        return self._parameter

    @property
    def persistent(self) -> bool:
        """
        Is true.

        ..  container:: example

            >>> abjad.BowPressure('overpressure').persistent
            True

        """
        return self._persistent

    @property
    def pressure(self) -> typing.Optional[str]:
        """
        Gets pressure of indicator.

        ..  container:: example

            Overpressure:

            >>> bow_pressure = abjad.BowPressure('overpressure')
            >>> bow_pressure.pressure
            'overpressure'

        ..  container:: example

            Underpressure:

            >>> bow_pressure = abjad.BowPressure('underpressure')
            >>> bow_pressure.pressure
            'underpressure'

        """
        return self._pressure

    @property
    def tweaks(self) -> None:
        """
        Are not implemented on bow pressure.
        """
        pass



from ..bundle import LilyPondFormatBundle
from ..storage import FormatSpecification, StorageFormatManager


class BarLine:
    r"""
    Bar line.

    ..  container:: example

        Final bar line:

        >>> staff = abjad.Staff("c'4 d'4 e'4 f'4")
        >>> bar_line = abjad.BarLine("|.")
        >>> abjad.attach(bar_line, staff[-1])
        >>> abjad.show(staff) # doctest: +SKIP

        >>> bar_line
        BarLine('|.', format_slot='after')

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'4
                d'4
                e'4
                f'4
                \bar "|."
            }

    ..  container:: example

        Specify repeat bars like this:

        >>> staff = abjad.Staff("c'4 d' e' f' g' a' b' c''")
        >>> abjad.attach(abjad.BarLine(".|:"), staff[3])
        >>> abjad.attach(abjad.BarLine(":|."), staff[-1])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'4
                d'4
                e'4
                f'4
                \bar ".|:"
                g'4
                a'4
                b'4
                c''4
                \bar ":|."
            }

        This allows you to bypass LilyPond's ``\volta`` command.

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_abbreviation", "_format_slot")

    _context = "Score"

    # scraped from LilyPond docs because LilyPond fails to error
    # on unrecognized string
    _known_abbreviations = (
        "",
        "|",
        ".",
        "||",
        ".|",
        "..",
        "|.|",
        "|.",
        ";",
        "!",
        ".|:",
        ":..:",
        ":|.|:",
        ":|.:",
        ":.|.:",
        "[|:",
        ":|][|:",
        ":|]",
        ":|.",
        "'",
    )

    ### INITIALIZER ##

    def __init__(self, abbreviation: str = "|", *, format_slot: str = "after") -> None:
        if abbreviation not in self._known_abbreviations:
            message = f"unknown bar line abbreviation: {repr(abbreviation)}\n"
            message += "Abbreviation must be one of these:\n"
            string = "\n    ".join([repr(_) for _ in self._known_abbreviations])
            message += string
            raise Exception(message)
        self._abbreviation = abbreviation
        assert isinstance(format_slot, str), repr(format_slot)
        self._format_slot = format_slot

    ### SPECIAL METHODS ###

    def __repr__(self):
        """
        Delegates to storage format manager.
        """
        return StorageFormatManager(self).get_repr_format()

    ### PRIVATE METHODS ###

    def _get_format_specification(self):
        return FormatSpecification(
            client=self,
            storage_format_is_indented=False,
            storage_format_args_values=[self.abbreviation],
        )

    def _get_lilypond_format(self):
        return rf'\bar "{self.abbreviation}"'

    def _get_lilypond_format_bundle(self, component=None):
        bundle = LilyPondFormatBundle()
        slot = bundle.get(self.format_slot)
        slot.commands.append(self._get_lilypond_format())
        return bundle

    ## PUBLIC PROPERTIES ##

    @property
    def abbreviation(self) -> str:
        r"""
        Gets abbreviation.

        ..  container:: example

            >>> bar_line = abjad.BarLine('|.')
            >>> bar_line.abbreviation
            '|.'

        ..  container:: example exception:

            Abbreviation error-checking looks like this:

            >>> abjad.BarLine("text")
            Traceback (most recent call last):
                ...
            Exception: unknown bar line abbreviation: 'text'
            Abbreviation must be one of these:
                ''
                '|'
                '.'
                '||'
                '.|'
                '..'
                '|.|'
                '|.'
                ';'
                '!'
                '.|:'
                ':..:'
                ':|.|:'
                ':|.:'
                ':.|.:'
                '[|:'
                ':|][|:'
                ':|]'
                ':|.'
                "'"

        """
        return self._abbreviation

    @property
    def context(self) -> str:
        r"""
        Gets (historically conventional) context.

        ..  container:: example

            >>> bar_line = abjad.BarLine('|.')
            >>> bar_line.context
            'Score'

        Override with ``abjad.attach(..., context='...')``.
        """
        return self._context

    @property
    def format_slot(self) -> str:
        r"""
        Gets format slot.

        ..  container:: example

            >>> abjad.BarLine("|").format_slot
            'after'

            >>> abjad.BarLine("|", format_slot="before").format_slot
            'before'

        ..  container:: example

            REGRESSION. You can attach a before barline and after barline to
            the same leaf:

            >>> staff = abjad.Staff("c'1 d'1")
            >>> bar_line_1 = abjad.BarLine(".|:", format_slot="before")
            >>> bar_line_2 = abjad.BarLine(":|.", format_slot="after")
            >>> assert bar_line_1.format_slot != bar_line_2.format_slot
            >>> abjad.attach(bar_line_1, staff[-1])
            >>> abjad.attach(bar_line_2, staff[-1])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(staff)
                >>> print(string)
                \new Staff
                {
                    c'1
                    \bar ".|:"
                    d'1
                    \bar ":|."
                }

        """
        return self._format_slot

    @property
    def tweaks(self) -> None:
        r"""
        Are not implemented on bar line.

        The LilyPond ``\bar`` command refuses tweaks.

        Use overrides instead.
        """
        pass



import collections
import importlib
import inspect
import types


class FormatSpecification:
    """
    Format specification.
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = "Storage formatting"

    __slots__ = (
        "_client",
        "_coerce_for_equality",
        "_repr_args_values",
        "_repr_is_bracketed",
        "_repr_is_indented",
        "_repr_keyword_names",
        "_repr_text",
        "_storage_format_args_values",
        "_storage_format_forced_override",
        "_storage_format_is_bracketed",
        "_storage_format_is_indented",
        "_storage_format_keyword_names",
        "_storage_format_text",
        "_template_names",
    )

    ### INITIALIZER ###

    def __init__(
        self,
        client=None,
        coerce_for_equality=None,
        repr_args_values=None,
        repr_is_bracketed=None,
        repr_is_indented=None,
        repr_keyword_names=None,
        repr_text=None,
        storage_format_args_values=None,
        storage_format_forced_override=None,
        storage_format_is_bracketed=None,
        storage_format_is_indented=True,
        storage_format_keyword_names=None,
        storage_format_text=None,
        template_names=None,
    ):
        self._client = client
        self._coerce_for_equality = self._coerce_boolean(coerce_for_equality)
        self._repr_args_values = self._coerce_tuple(repr_args_values)
        self._repr_is_bracketed = self._coerce_boolean(repr_is_bracketed)
        self._repr_is_indented = self._coerce_boolean(repr_is_indented)
        self._repr_keyword_names = self._coerce_tuple(repr_keyword_names)
        self._repr_text = self._coerce_string(repr_text)
        self._storage_format_args_values = self._coerce_tuple(
            storage_format_args_values
        )
        self._storage_format_forced_override = storage_format_forced_override
        self._storage_format_is_bracketed = self._coerce_boolean(
            storage_format_is_bracketed
        )
        self._storage_format_is_indented = self._coerce_boolean(
            storage_format_is_indented
        )
        self._storage_format_keyword_names = self._coerce_tuple(
            storage_format_keyword_names
        )
        self._storage_format_text = self._coerce_string(storage_format_text)
        self._template_names = self._coerce_tuple(template_names)

    ### PRIVATE METHODS ###

    def _coerce_boolean(self, value):
        if value is not None:
            return bool(value)

    def _coerce_string(self, value):
        if value is not None:
            return str(value)

    def _coerce_tuple(self, value):
        if value is not None:
            return tuple(value)

    ### PUBLIC PROPERTIES ###

    @property
    def client(self):
        return self._client

    @property
    def coerce_for_equality(self):
        return self._coerce_for_equality

    @property
    def repr_args_values(self):
        return self._repr_args_values

    @property
    def repr_is_bracketed(self):
        return self._repr_is_bracketed

    @property
    def repr_is_indented(self):
        return self._repr_is_indented

    @property
    def repr_keyword_names(self):
        return self._repr_keyword_names

    @property
    def repr_text(self):
        return self._repr_text

    @property
    def storage_format_args_values(self):
        return self._storage_format_args_values

    @property
    def storage_format_forced_override(self):
        return self._storage_format_forced_override

    @property
    def storage_format_is_bracketed(self):
        return self._storage_format_is_bracketed

    @property
    def storage_format_is_indented(self):
        return self._storage_format_is_indented

    @property
    def storage_format_keyword_names(self):
        return self._storage_format_keyword_names

    @property
    def storage_format_text(self):
        return self._storage_format_text

    @property
    def template_names(self):
        return self._template_names


class StorageFormatSpecification:
    """
    Storage format specification.
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = "Storage formatting"

    __slots__ = (
        "_repr_text",
        "_include_abjad_namespace",
        "_instance",
        "_is_bracketed",
        "_is_indented",
        "_keyword_argument_names",
        "_positional_argument_values",
        "_storage_format_text",
    )

    ### INITIALIZER ###

    def __init__(
        self,
        instance=None,
        repr_text=None,
        include_abjad_namespace=None,
        is_bracketed=False,
        is_indented=True,
        keyword_argument_names=None,
        positional_argument_values=None,
        storage_format_text=None,
    ):
        self._instance = instance

        if repr_text is not None:
            repr_text = str(repr_text)
        self._repr_text = repr_text

        self._include_abjad_namespace = bool(include_abjad_namespace)
        self._is_bracketed = bool(is_bracketed)
        self._is_indented = bool(is_indented)

        if keyword_argument_names is not None:
            keyword_argument_names = tuple(keyword_argument_names)
        self._keyword_argument_names = keyword_argument_names

        if positional_argument_values is not None:
            positional_argument_values = tuple(positional_argument_values)
        self._positional_argument_values = positional_argument_values

        if storage_format_text is not None:
            storage_format_text = str(storage_format_text)
        self._storage_format_text = storage_format_text

    ### SPECIAL METHODS ###

    def __repr__(self) -> str:
        """
        Gets interpreter representation.
        """
        return StorageFormatManager(self).get_repr_format()

    ### PUBLIC PROPERTIES ###

    @property
    def include_abjad_namespace(self):
        """
        Is true when storage specification includes Abjad namespace.

        Returns true or false.
        """
        return self._include_abjad_namespace

    @property
    def instance(self):
        """
        Gets instance of storage specification.

        Returns string.
        """
        return self._instance

    @property
    def is_bracketed(self):
        """
        Is true when storage specification is bracketed.

        Returns true or false.
        """
        return self._is_bracketed

    @property
    def is_indented(self):
        """
        Is true when storage format is indented.

        Returns true or false.
        """
        return self._is_indented

    @property
    def keyword_argument_names(self):
        """
        Gets keyword argument names of storage format.

        Returns tuple.
        """
        return self._keyword_argument_names

    @property
    def positional_argument_values(self):
        """
        Gets positional argument values.

        Returns tuple.
        """
        return self._positional_argument_values

    @property
    def repr_text(self):
        """
        Gets interpreter representation of storage specification.

        Returns string.
        """
        return self._repr_text

    @property
    def storage_format_text(self):
        """
        Gets storage format text.

        Returns tuple.
        """
        return self._storage_format_text


class StorageFormatManager:
    """
    Manages Abjad object storage formats.
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = "Storage formatting"

    __slots__ = (
        "_client",
        "_format_specification",
        "_signature_accepts_args",
        "_signature_accepts_keywords",
        "_signature_keyword_names",
        "_signature_positional_names",
    )

    _exclude_tools_package = (
        "core",
        "indicators",
        "markup",
        "math",
        "meter",
        "pitch",
        "scheme",
        "timespans",
    )

    _unindented_whitespace = "", "", ", "

    _indented_whitespace = "    ", "\n", ",\n"

    ### INITIALIZER ###

    def __init__(self, client=None):
        self._client = client
        self._format_specification = None
        (
            self._signature_positional_names,
            self._signature_keyword_names,
            self._signature_accepts_args,
            self._signature_accepts_keywords,
        ) = self.inspect_signature(self._client)

    ### PRIVATE METHODS ###

    def _dispatch_formatting(self, as_storage_format=True, is_indented=True):
        if isinstance(self._client, types.MethodType):
            return self._format_method()
        elif isinstance(self._client, type):
            return self._format_class(as_storage_format, is_indented)
        elif as_storage_format and (
            hasattr(self._client, "_get_storage_format_specification")
            or hasattr(self._client, "_get_format_specification")
        ):
            pieces = self._format_specced_object(as_storage_format=as_storage_format)
            return list(pieces)
        elif not as_storage_format and hasattr(
            self._client, "_get_format_specification"
        ):
            pieces = self._format_specced_object(as_storage_format=as_storage_format)
            return list(pieces)
        elif isinstance(self._client, (list, tuple)):
            return self._format_sequence(as_storage_format, is_indented)
        elif isinstance(self._client, collections.OrderedDict):
            return self._format_ordered_mapping(as_storage_format, is_indented)
        elif hasattr(self._client, "_collection") and isinstance(
            self._client._collection, collections.OrderedDict
        ):
            return self._format_ordered_mapping(as_storage_format, is_indented)
        elif isinstance(self._client, dict):
            return self._format_mapping(as_storage_format, is_indented)
        elif isinstance(self._client, float):
            return repr(round(self._client, 15)).split("\n")
        return repr(self._client).split("\n")

    def _format_class(self, as_storage_format, is_indented):
        if as_storage_format:
            root_package_name = self.get_root_package_name()
            root_package = importlib.import_module(root_package_name)
            parts = [root_package_name]
            if self._client.__name__ not in dir(root_package):
                tools_package_name = self.get_tools_package_name()
                parts.append(tools_package_name)
            parts.append(self._client.__name__)
            result = ".".join(parts)
        else:
            result = self._client.__name__
        return [result]

    def _format_mapping(self, as_storage_format, is_indented):
        result = []
        prefix, infix, suffix = self._get_whitespace(is_indented)
        result.append("{" + infix)
        for key, value in sorted(self._client.items()):
            key_agent = type(self)(key)
            key_pieces = key_agent._dispatch_formatting(
                as_storage_format=as_storage_format, is_indented=is_indented
            )
            value_agent = type(self)(value)
            value_pieces = value_agent._dispatch_formatting(
                as_storage_format=as_storage_format, is_indented=is_indented
            )
            for line in key_pieces[:-1]:
                result.append(prefix + line)
            result.append(f"{prefix}{key_pieces[-1]}: {value_pieces[0]}")
            for line in value_pieces[1:]:
                result.append(prefix + line)
            result[-1] = result[-1] + suffix
        if not is_indented:
            result[-1] = result[-1].rstrip(suffix) + infix
        result.append(prefix + "}")
        return result

    def _format_method(self, as_storage_format, is_indented):
        return []

    def _format_ordered_mapping(self, as_storage_format, is_indented):
        result = []
        prefix, infix, suffix = self._get_whitespace(is_indented)
        result.append("[" + infix)
        for item in list(self._client.items()):
            item_agent = type(self)(item)
            item_pieces = item_agent._dispatch_formatting(
                as_storage_format=as_storage_format, is_indented=is_indented
            )
            for line in item_pieces:
                result.append(prefix + line)
            result[-1] = result[-1] + suffix
        if not is_indented:
            result[-1] = result[-1].rstrip(suffix) + infix
        result.append(prefix + "]")
        return result

    def _format_sequence(self, as_storage_format, is_indented):
        result = []
        prefix, infix, suffix = self._get_whitespace(is_indented)
        # just return the repr, if all contents are builtin types
        prototype = (bool, int, float, str, type(None))
        if all(isinstance(x, prototype) for x in self._client):
            piece = repr(self._client)
            if len(piece) < 50:
                return [repr(self._client)]
        if isinstance(self._client, list):
            braces = "[", "]"
        else:
            braces = "(", ")"
        result.append(braces[0] + infix)
        for x in self._client:
            agent = type(self)(x)
            pieces = agent._dispatch_formatting(
                as_storage_format=as_storage_format, is_indented=is_indented
            )
            for line in pieces[:-1]:
                result.append(prefix + line)
            result.append(prefix + pieces[-1] + suffix)
        if not is_indented:
            if isinstance(self._client, list) or 1 < len(self._client):
                result[-1] = result[-1].rstrip(suffix)
            else:
                result[-1] = result[-1].rstrip()
        result.append(prefix + braces[1])
        return result

    def _format_specced_object(self, as_storage_format=True):
        if hasattr(self._client, "_get_format_specification"):
            specification = self._client._get_format_specification()
            if specification.storage_format_forced_override is not None:
                return [specification.storage_format_forced_override]
        formatting_keywords = self._get_formatting_keywords(as_storage_format)
        args_values = formatting_keywords["args_values"]
        as_storage_format = formatting_keywords["as_storage_format"]
        is_bracketed = formatting_keywords["is_bracketed"]
        is_indented = formatting_keywords["is_indented"]
        keyword_names = formatting_keywords["keyword_names"]
        text = formatting_keywords["text"]
        result = []
        if is_bracketed:
            result.append("<")
        if text is not None:
            result.append(text)
        else:
            prefix, infix, suffix = self._get_whitespace(is_indented)
            class_name_prefix = self.get_class_name_prefix(as_storage_format)
            positional_argument_pieces = []
            for value in args_values:
                agent = type(self)(value)
                pieces = agent._dispatch_formatting(
                    as_storage_format=as_storage_format,
                    is_indented=is_indented,
                )
                for piece in pieces[:-1]:
                    positional_argument_pieces.append(prefix + piece)
                positional_argument_pieces.append(prefix + pieces[-1] + suffix)
            keyword_argument_pieces = []
            for name in keyword_names:
                value = self._get(name)
                if value is None or isinstance(value, types.MethodType):
                    continue
                agent = type(self)(value)
                pieces = agent._dispatch_formatting(
                    as_storage_format=as_storage_format,
                    is_indented=is_indented,
                )
                pieces[0] = f"{name}={pieces[0]}"
                for piece in pieces[:-1]:
                    keyword_argument_pieces.append(prefix + piece)
                keyword_argument_pieces.append(prefix + pieces[-1] + suffix)
            if not positional_argument_pieces and not keyword_argument_pieces:
                result.append(f"{class_name_prefix}()")
            else:
                result.append(f"{class_name_prefix}({infix}")
                result.extend(positional_argument_pieces)
                if positional_argument_pieces and not keyword_argument_pieces:
                    result[-1] = result[-1].rstrip(suffix) + infix
                else:
                    result.extend(keyword_argument_pieces)
                if not as_storage_format:
                    result[-1] = result[-1].rstrip(suffix) + infix
                if is_indented:
                    result.append(f"{prefix})")
                else:
                    result.append(")")
        if is_bracketed:
            result.append(">")
        if not is_indented:
            return ["".join(result)]
        return result

    def _get(self, name):
        value = None
        try:
            value = getattr(self._client, name, None)
            if value is None:
                value = getattr(self._client, "_" + name, None)
            if value is None:
                value = getattr(self._client, "_" + name.rstrip("_"), None)
        except AttributeError:
            try:
                value = self._client[name]
            except (TypeError, KeyError):
                value = None
        return value

    def _get_formatting_keywords(self, as_storage_format=True):
        # NOTE: This acts to abstract-away our competing spec-specs.
        #       It can probably be removed/reduced in the near future.
        if as_storage_format:
            spec = None
            if hasattr(self._client, "_get_storage_format_specification"):
                spec = self._client._get_storage_format_specification()
            if spec:
                # print('STORAGE', type(self._client), getattr(self._client, 'name', None))
                via = "_get_storage_format_specification()"
                args_values = spec.positional_argument_values
                is_bracketed = False
                is_indented = spec.is_indented
                keyword_names = spec.keyword_argument_names
                text = spec.storage_format_text
            else:
                spec = self.format_specification
                via = "_get_format_specification()"
                args_values = spec.storage_format_args_values
                is_bracketed = spec.storage_format_is_bracketed
                is_indented = spec.storage_format_is_indented
                keyword_names = spec.storage_format_keyword_names
                text = spec.storage_format_text
        else:
            spec = self.format_specification
            via = "_get_format_specification()"
            args_values = spec.repr_args_values
            if args_values is None:
                args_values = spec.storage_format_args_values
            is_bracketed = spec.repr_is_bracketed
            is_indented = spec.repr_is_indented
            keyword_names = spec.repr_keyword_names
            if keyword_names is None:
                keyword_names = spec.storage_format_keyword_names
            text = spec.repr_text
            if text is None:
                text = spec.storage_format_text
        if keyword_names is None:
            keyword_names = self.signature_keyword_names
        if args_values is None:
            args_values = tuple(self._get(_) for _ in self.signature_positional_names)
        if args_values:
            keyword_names = list(keyword_names)
            names = self.signature_positional_names
            if not self.signature_accepts_args:
                names += self.signature_keyword_names
            names = names[: len(args_values)]
            for name in names:
                if name in keyword_names:
                    keyword_names.remove(name)
            keyword_names = tuple(keyword_names)
        return dict(
            args_values=args_values,
            as_storage_format=as_storage_format,
            is_bracketed=is_bracketed,
            is_indented=is_indented,
            keyword_names=keyword_names,
            text=text,
            via=via,
        )

    @staticmethod
    def _get_module_path_parts(subject):
        if isinstance(subject, type):
            class_ = subject
        elif type(subject) is subject.__class__:
            class_ = type(subject)
        class_name = class_.__name__
        parts = class_.__module__.split(".")
        while parts and parts[-1] == class_name:
            parts.pop()
        parts.append(class_name)
        return parts

    def _get_whitespace(self, is_indented):
        if is_indented:
            return self._indented_whitespace
        return self._unindented_whitespace

    @staticmethod
    def _make_hashable(value):
        if isinstance(value, dict):
            value = tuple(value.items())
        elif isinstance(value, list):
            value = tuple(value)
        elif isinstance(value, (set, frozenset)):
            value = tuple(value)
        return value

    def _map_positional_values_to_names(self, values):
        names = self.signature_positional_names
        if not self.signature_accepts_args:
            names += self.signature_keyword_names
        names = names[: len(values)]
        return names

    ### PUBLIC PROPERTIES ###

    @property
    def client(self):
        return self._client

    @property
    def format_specification(self):
        if self._format_specification is None:
            if not isinstance(self._client, type) and hasattr(
                self._client, "_get_format_specification"
            ):
                self._format_specification = self._client._get_format_specification()
            else:
                self._format_specification = FormatSpecification(self._client)
        return self._format_specification

    @property
    def signature_accepts_args(self):
        return self._signature_accepts_args

    @property
    def signature_accepts_keywords(self):
        return self._signature_accepts_keywords

    @property
    def signature_keyword_names(self):
        return self._signature_keyword_names

    @property
    def signature_names(self):
        return self.signature_positional_names + self.signature_keyword_names

    @property
    def signature_positional_names(self):
        return self._signature_positional_names

    ### PUBLIC METHODS ###

    @staticmethod
    def compare_objects(object_one, object_two) -> bool:
        """
        Compares ``object_one`` to ``object_two``.
        """
        manager_one = StorageFormatManager(object_one)
        if manager_one.format_specification.coerce_for_equality:
            try:
                object_two = type(object_one)(object_two)
            except (TypeError, ValueError):
                return False
        elif not isinstance(object_two, type(object_one)):
            return False
        manager_two = StorageFormatManager(object_two)
        template_1 = manager_one.get_template_dict()
        template_2 = manager_two.get_template_dict()
        return template_1 == template_2

    def get_class_name_prefix(
        self, as_storage_format, include_root_package=None
    ) -> str:
        """
        Gets class name prefix.
        """
        if not isinstance(self._client, type):
            class_name = type(self._client).__name__
        else:
            class_name = self._client.__name__
        if as_storage_format:
            root_package_name = self.get_root_package_name()
            root_package = importlib.import_module(root_package_name)
            parts = []
            if root_package_name != "abjadext":
                parts.append(root_package_name)
            if class_name not in dir(root_package):
                name = StorageFormatManager(self._client).get_tools_package_name()
                parts.append(name)
            parts.append(class_name)
            return ".".join(parts)
        return class_name

    def get_hash_values(self):
        """
        Gets hash values.
        """
        values = []
        if isinstance(self._client, type):
            values.append(self._client)
        else:
            values.append(type(self._client))
        template_items = sorted(self.get_template_dict().items())
        values.extend(self._make_hashable(v) for k, v in template_items)
        return tuple(values)

    def get_repr_format(self):
        """
        Gets repr format.
        """
        pieces = self._format_specced_object(as_storage_format=False)
        return "".join(pieces)

    def get_repr_keyword_dict(self):
        """
        Gets repr keyword dictionary.
        """
        names = self.specification.repr_keyword_names
        if names is None:
            specification = StorageFormatSpecification(self.client)
            names = specification.keyword_argument_names or ()
        keyword_dict = {}
        for name in names:
            keyword_dict[name] = self._get(name)
        return keyword_dict

    def get_repr_positional_values(self):
        """
        Gets repr positional values.
        """
        values = self.specification.repr_args_values
        if values is None:
            specification = StorageFormatSpecification(self.client)
            values = specification.positional_argument_values or ()
        return tuple(values)

    def get_root_package_name(self) -> str:
        """
        Gets root package name.
        """
        return self._get_module_path_parts(self._client)[0]

    def get_storage_format(self) -> str:
        """
        Gets storage format.
        """
        pieces = self._format_specced_object(as_storage_format=True)
        result = "".join(pieces)
        return result

    def get_storage_format_keyword_dict(self):
        """
        Gets storage format keyword dictionary.
        """
        names = self.specification.storage_format_keyword_names
        if names is None:
            if hasattr(self.client, "_get_storage_format_specification"):
                specification = self.client._get_storage_format_specification()
            else:
                specification = StorageFormatSpecification(self.client)
            names = specification.keyword_argument_names or ()
        keyword_dict = {}
        for name in names:
            keyword_dict[name] = self._get(name)
        return keyword_dict

    def get_storage_format_positional_values(self):
        """
        Gets storage format positional values.
        """
        values = self.specification.storage_format_args_values
        if values is None:
            if hasattr(self.client, "_get_storage_format_specification"):
                specification = self.client._get_storage_format_specification()
            else:
                specification = StorageFormatSpecification(self.client)
            values = specification.positional_argument_values or ()
        return tuple(values)

    def get_template_dict(self):
        """
        Gets template dictionary.
        """
        template_names = self.format_specification.template_names
        if template_names is None:
            template_names = self.signature_names
            # TODO: This will be factored out when SFS/SFM are removed.
            if hasattr(self.client, "_get_storage_format_specification"):
                specification = self.client._get_storage_format_specification()
                template_names.extend(specification._keyword_argument_names or ())
            else:
                template_names.extend(
                    self.format_specification.storage_format_keyword_names or ()
                )
            template_names = sorted(set(template_names))
        template_dict = collections.OrderedDict()
        for name in template_names:
            template_dict[name] = self._get(name)
        return template_dict

    def get_tools_package_name(self):
        """
        Gets tools package name.
        """
        parts = self._get_module_path_parts(self._client)
        if parts[0] == "abjadext":
            return parts[1]
        if parts[0] in ("abjad", "ide"):
            for part in reversed(parts):
                if part == parts[-1]:
                    continue
                return part
        return ".".join(parts[:-1])

    @classmethod
    def inspect_signature(class_, subject):
        """
        Inspects signature of ``subject``.
        """
        positional_names = []
        keyword_names = []
        accepts_args = False
        accepts_keywords = False
        if not isinstance(subject, type):
            subject = type(subject)
        try:
            signature = inspect.signature(subject)
        except ValueError:
            return (
                positional_names,
                keyword_names,
                accepts_args,
                accepts_keywords,
            )
        for name, parameter in signature.parameters.items():
            if parameter.kind == inspect._POSITIONAL_OR_KEYWORD:
                if parameter.default == parameter.empty:
                    positional_names.append(name)
                else:
                    keyword_names.append(name)
            # Python 3 allow keyword only parameters:
            elif (
                hasattr(inspect, "_KEYWORD_ONLY")
                and parameter.kind == inspect._KEYWORD_ONLY
            ):
                keyword_names.append(name)
            elif parameter.kind == inspect._VAR_POSITIONAL:
                accepts_args = True
            elif parameter.kind == inspect._VAR_KEYWORD:
                accepts_keywords = True
        return (positional_names, keyword_names, accepts_args, accepts_keywords)


### FUNCTIONS ###


def storage(argument):
    """
    Gets storage format of ``argument``.
    """
    return StorageFormatManager(argument).get_storage_format()



import math
import re
import typing

import quicktions

from . import exceptions
from . import math as _math
from . import storage


class Duration(quicktions.Fraction):
    """
    Duration.

    ..  container:: example

        Initializes from integer numerator:

        >>> abjad.Duration(3)
        Duration(3, 1)

    ..  container:: example

        Initializes from integer numerator and denominator:

        >>> abjad.Duration(3, 16)
        Duration(3, 16)

    ..  container:: example

        Initializes from integer-equivalent numeric numerator:

        >>> abjad.Duration(3.0)
        Duration(3, 1)

    ..  container:: example

        Initializes from integer-equivalent numeric numerator and denominator:

        >>> abjad.Duration(3.0, 16)
        Duration(3, 16)

    ..  container:: example

        Initializes from integer-equivalent singleton:

        >>> abjad.Duration((3,))
        Duration(3, 1)

    ..  container:: example

        Initializes from integer-equivalent pair:

        >>> abjad.Duration((3, 16))
        Duration(3, 16)

    ..  container:: example

        Initializes from other duration:

        >>> abjad.Duration(abjad.Duration(3, 16))
        Duration(3, 16)

    ..  container:: example

        Intializes from fraction:

        >>> import quicktions
        >>> abjad.Duration(quicktions.Fraction(3, 16))
        Duration(3, 16)

    ..  container:: example

        Initializes from solidus string:

        >>> abjad.Duration('3/16')
        Duration(3, 16)

    ..  container:: example

        Initializes from nonreduced fraction:

        >>> abjad.Duration(abjad.NonreducedFraction(6, 32))
        Duration(3, 16)

    ..  container:: example

        Durations inherit from built-in fraction:

        >>> isinstance(abjad.Duration(3, 16), quicktions.Fraction)
        True

    ..  container:: example

        Durations are numbers:

        >>> import numbers

        >>> isinstance(abjad.Duration(3, 16), numbers.Number)
        True

    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### CONSTRUCTOR ###

    def __new__(class_, *arguments):
        """
        Makes new duration.
        """
        if len(arguments) == 1:
            argument = arguments[0]
            if type(argument) is class_:
                return argument
            if isinstance(argument, NonreducedFraction):
                return quicktions.Fraction.__new__(class_, *argument.pair)
            try:
                return quicktions.Fraction.__new__(class_, *argument)
            except (AttributeError, TypeError):
                pass
            try:
                return quicktions.Fraction.__new__(class_, argument)
            except (AttributeError, TypeError):
                pass
            if (
                isinstance(argument, tuple)
                and len(argument) == 2
                and _math.is_integer_equivalent(argument[0])
                and _math.is_integer_equivalent(argument[1])
                and not argument[1] == 0
            ):
                return quicktions.Fraction.__new__(
                    class_, int(argument[0]), int(argument[1])
                )
            try:
                return quicktions.Fraction.__new__(class_, argument.duration)
            except AttributeError:
                pass
            if isinstance(argument, str) and "/" not in argument:
                result = Duration._initialize_from_lilypond_duration_string(argument)
                return quicktions.Fraction.__new__(class_, result)
            if (
                isinstance(argument, tuple)
                and len(argument) == 1
                and _math.is_integer_equivalent(argument[0])
            ):
                return quicktions.Fraction.__new__(class_, int(argument[0]))
        else:
            try:
                return quicktions.Fraction.__new__(class_, *arguments)
            except TypeError:
                pass
            if _math.all_are_integer_equivalent_numbers(arguments):
                return quicktions.Fraction.__new__(class_, *[int(x) for x in arguments])
        raise ValueError(f"can not construct duration: {arguments!r}.")

    ### SPECIAL METHODS ###

    def __abs__(self, *arguments):
        """
        Gets absolute value of duration.

        Returns nonnegative duration.
        """
        return type(self)(quicktions.Fraction.__abs__(self, *arguments))

    def __add__(self, *arguments):
        """
        Adds duration to ``arguments``.

        ..  container:: example

            Returns duration when ``arguments`` is a duration:

            >>> duration_1 = abjad.Duration(1, 2)
            >>> duration_2 = abjad.Duration(3, 2)
            >>> duration_1 + duration_2
            Duration(2, 1)

        ..  container:: example

            Returns nonreduced fraction when ``arguments`` is a nonreduced
            fraction:

            >>> duration = abjad.Duration(1, 2)
            >>> nonreduced_fraction = abjad.NonreducedFraction(3, 6)
            >>> duration + nonreduced_fraction
            NonreducedFraction(6, 6)

        Returns duration.
        """
        if len(arguments) == 1 and isinstance(arguments[0], NonreducedFraction):
            result = arguments[0].__radd__(self)
        else:
            result = type(self)(quicktions.Fraction.__add__(self, *arguments))
        return result

    def __div__(self, *arguments):
        """
        Divides duration by ``arguments``.

        ..  container:: example

            >>> abjad.Duration(1) / abjad.NonreducedFraction(3, 3)
            NonreducedFraction(3, 3)


        >>> abjad.NonreducedFraction(3, 3) / abjad.Duration(1)
        NonreducedFraction(3, 3)

        Returns multiplier.
        """
        if len(arguments) == 1 and isinstance(arguments[0], type(self)):
            fraction = quicktions.Fraction.__truediv__(self, *arguments)
            result = Multiplier(fraction)
        elif len(arguments) == 1 and isinstance(arguments[0], NonreducedFraction):
            result = arguments[0].__rdiv__(self)
        else:
            result = type(self)(quicktions.Fraction.__truediv__(self, *arguments))
        return result

    def __divmod__(self, *arguments):
        """
        Equals the pair (duration // ``arguments``, duration % ``arguments``).

        Returns pair.
        """
        truncated, residue = quicktions.Fraction.__divmod__(self, *arguments)
        truncated = type(self)(truncated)
        residue = type(self)(residue)
        return truncated, residue

    def __eq__(self, argument) -> bool:
        """
        Is true when duration equals ``argument``.
        """
        return quicktions.Fraction.__eq__(self, argument)

    def __ge__(self, argument) -> bool:
        """
        Is true when duration is greater than or equal to ``argument``.
        """
        return quicktions.Fraction.__ge__(self, argument)

    def __gt__(self, argument) -> bool:
        """
        Is true when duration is greater than ``argument``.
        """
        return quicktions.Fraction.__gt__(self, argument)

    def __hash__(self) -> int:
        """
        Hashes duration.

        Required to be explicitly redefined on Python 3 if __eq__ changes.
        """
        return super().__hash__()

    def __le__(self, argument) -> bool:
        """
        Is true when duration is less than or equal to ``argument``.
        """
        return quicktions.Fraction.__le__(self, argument)

    def __lt__(self, argument) -> bool:
        """
        Is true when duration is less than ``argument``.
        """
        return quicktions.Fraction.__lt__(self, argument)

    def __mod__(self, *arguments):
        """
        Modulus operator applied to duration.

        Returns duration.
        """
        return type(self)(quicktions.Fraction.__mod__(self, *arguments))

    def __mul__(self, *arguments):
        """
        Duration multiplied by ``arguments``.

        ..  container:: example

            Returns a new duration when ``arguments`` is a duration:

            >>> duration_1 = abjad.Duration(1, 2)
            >>> duration_2 = abjad.Duration(3, 2)
            >>> duration_1 * duration_2
            Duration(3, 4)

        ..  container:: example

            Returns nonreduced fraction when ``arguments`` is a nonreduced
            fraction:

            >>> duration = abjad.Duration(1, 2)
            >>> nonreduced_fraction = abjad.NonreducedFraction(3, 6)
            >>> duration * nonreduced_fraction
            NonreducedFraction(3, 12)

        Returns duration or nonreduced fraction.
        """
        if len(arguments) == 1 and isinstance(arguments[0], NonreducedFraction):
            result = arguments[0].__rmul__(self)
        else:
            result = type(self)(quicktions.Fraction.__mul__(self, *arguments))
        return result

    def __ne__(self, argument) -> bool:
        """
        Redefined explicitly because ``abjad.Duration`` inherits from built-in
        ``quicktions.Fraction``:

        "The __ne__ method follows automatically from __eq__ only if __ne__
        isn't already defined in a superclass.  So, if you're inheriting from a
        builtin, it's best to override both."

        See https://bugs.python.org/issue4395#msg89533.

        ..  container:: example

            REGRESSION:

            >>> offset_1 = abjad.Offset(1)
            >>> offset_2 = abjad.Offset(1, displacement=(-1, 16))

            >>> offset_1 == offset_2
            False

            >>> offset_1 != offset_2
            True

        """
        return not self == argument

    def __neg__(self, *arguments):
        """
        Negates duration.

        Returns new duration.
        """
        return type(self)(quicktions.Fraction.__neg__(self, *arguments))

    def __pos__(self, *arguments):
        """
        Get positive duration.

        Returns new duration.
        """
        return type(self)(quicktions.Fraction.__pos__(self, *arguments))

    def __pow__(self, *arguments):
        """
        Raises duration to ``arguments`` power.

        Returns new duration.
        """
        return type(self)(quicktions.Fraction.__pow__(self, *arguments))

    def __radd__(self, *arguments):
        """
        Adds ``arguments`` to duration.

        Returns new duration.
        """
        return type(self)(quicktions.Fraction.__radd__(self, *arguments))

    def __rdiv__(self, *arguments):
        """
        Divides ``arguments`` by duration.

        Returns new duration.
        """
        return type(self)(quicktions.Fraction.__rdiv__(self, *arguments))

    def __rdivmod__(self, *arguments):
        """
        Documentation required.
        """
        return type(self)(quicktions.Fraction.__rdivmod__(self, *arguments))

    def __reduce__(self):
        """
        Documentation required.
        """
        return type(self), (self.numerator, self.denominator)

    def __reduce_ex__(self, protocol):
        """
        Documentation required.
        """
        return type(self), (self.numerator, self.denominator)

    def __repr__(self) -> str:
        """
        Gets interpreter representation.
        """
        return storage.StorageFormatManager(self).get_repr_format()

    def __rmod__(self, *arguments):
        """
        Documentation required.
        """
        return type(self)(quicktions.Fraction.__rmod__(self, *arguments))

    def __rmul__(self, *arguments):
        """
        Multiplies ``arguments`` by duration.

        Returns new duration.
        """
        return type(self)(quicktions.Fraction.__rmul__(self, *arguments))

    def __rpow__(self, *arguments):
        """
        Raises ``arguments`` to the power of duration.

        Returns new duration.
        """
        return type(self)(quicktions.Fraction.__rpow__(self, *arguments))

    def __rsub__(self, *arguments):
        """
        Subtracts duration from ``arguments``.

        Returns new duration.
        """
        return type(self)(quicktions.Fraction.__rsub__(self, *arguments))

    def __rtruediv__(self, *arguments):
        """
        Documentation required.

        Returns new duration.
        """
        return type(self)(quicktions.Fraction.__rtruediv__(self, *arguments))

    def __sub__(self, *arguments):
        """
        Subtracts ``arguments`` from duration.

        ..  container:: example

            >>> abjad.Duration(1, 2) - abjad.NonreducedFraction(2, 8)
            NonreducedFraction(2, 8)

            >>> abjad.NonreducedFraction(4, 8) - abjad.Duration(1, 4)
            NonreducedFraction(2, 8)

        Returns new duration.
        """
        if len(arguments) == 1 and isinstance(arguments[0], NonreducedFraction):
            return arguments[0].__rsub__(self)
        else:
            return type(self)(quicktions.Fraction.__sub__(self, *arguments))

    def __truediv__(self, *arguments):
        """
        Documentation required.
        """
        return self.__div__(*arguments)

    ### PRIVATE METHODS ###

    def _get_format_specification(self):
        return storage.FormatSpecification(
            client=self,
            storage_format_args_values=[self.numerator, self.denominator],
            storage_format_is_indented=False,
            storage_format_keyword_names=[],
        )

    @staticmethod
    def _group_by_implied_prolation(durations):
        durations = [NonreducedFraction(duration) for duration in durations]
        assert 0 < len(durations)
        group = [durations[0]]
        result = [group]
        for d in durations[1:]:
            d_f = set(_math.factors(d.denominator))
            d_f.discard(2)
            gd_f = set(_math.factors(group[0].denominator))
            gd_f.discard(2)
            if d_f == gd_f:
                group.append(d)
            else:
                group = [d]
                result.append(group)
        return result

    @staticmethod
    def _initialize_from_lilypond_duration_string(duration_string):
        numeric_body_strings = [str(2 ** n) for n in range(8)]
        other_body_strings = [r"\\breve", r"\\longa", r"\\maxima"]
        body_strings = numeric_body_strings + other_body_strings
        body_strings = "|".join(body_strings)
        pattern = r"^(%s)(\.*)$" % body_strings
        match = re.match(pattern, duration_string)
        if match is None:
            raise TypeError(f"incorrect duration string format: {duration_string!r}.")
        body_string, dots_string = match.groups()
        try:
            body_denominator = int(body_string)
            body_duration = quicktions.Fraction(1, body_denominator)
        except ValueError:
            if body_string == r"\breve":
                body_duration = quicktions.Fraction(2)
            elif body_string == r"\longa":
                body_duration = quicktions.Fraction(4)
            elif body_string == r"\maxima":
                body_duration = quicktions.Fraction(8)
            else:
                raise ValueError(f"unknown body string: {body_string!r}.")
        rational = body_duration
        for n in range(len(dots_string)):
            exponent = n + 1
            denominator = 2 ** exponent
            multiplier = quicktions.Fraction(1, denominator)
            addend = multiplier * body_duration
            rational += addend
        return rational

    @staticmethod
    def _least_power_of_two_greater_equal(n, i=0) -> int:
        """
        When ``i = 2``, returns the second integer power of 2 greater than
        the least integer power of 2 greater than or equal to ``n``, and, in
        general, return the ``i`` th integer power of 2 greater than the least
        integer power of 2 greater than or equal to ``n``.
        """
        assert isinstance(n, (int, float, quicktions.Fraction)), repr(n)
        assert 0 <= n, repr(n)
        result = 2 ** (int(math.ceil(math.log(n, 2))) + i)
        return result

    ### PUBLIC PROPERTIES ###

    @property
    def dot_count(self) -> int:
        r"""
        Gets dot count.

        ..  container:: example

            Gets dot count:

            >>> for n in range(1, 16 + 1):
            ...     try:
            ...         duration = abjad.Duration(n, 16)
            ...         sixteenths = duration.with_denominator(16)
            ...         dot_count = duration.dot_count
            ...         string = f"{sixteenths!s}\t{dot_count}"
            ...         print(string)
            ...     except abjad.AssignabilityError:
            ...         sixteenths = duration.with_denominator(16)
            ...         print(f"{sixteenths!s}\t--")
            ...
            1/16    0
            2/16    0
            3/16    1
            4/16    0
            5/16    --
            6/16    1
            7/16    2
            8/16    0
            9/16    --
            10/16   --
            11/16   --
            12/16   1
            13/16   --
            14/16   2
            15/16   3
            16/16   0

        Dot count defined equal to number of dots required to notate duration.

        Raises assignability error when duration is not assignable.
        """
        if not self.is_assignable:
            raise exceptions.AssignabilityError
        binary_string = _math.integer_to_binary_string(self.numerator)
        digit_sum = sum([int(x) for x in list(binary_string)])
        dot_count = digit_sum - 1
        return dot_count

    @property
    def equal_or_greater_assignable(self) -> "Duration":
        r"""
        Gets assignable duration equal to or just greater than this
        duration.

        ..  container:: example

            Gets equal-or-greater assignable duration:

            >>> for numerator in range(1, 16 + 1):
            ...     duration = abjad.Duration(numerator, 16)
            ...     result = duration.equal_or_greater_assignable
            ...     sixteenths = duration.with_denominator(16)
            ...     print(f"{sixteenths!s}\t{result!s}")
            ...
            1/16    1/16
            2/16    1/8
            3/16    3/16
            4/16    1/4
            5/16    3/8
            6/16    3/8
            7/16    7/16
            8/16    1/2
            9/16    3/4
            10/16   3/4
            11/16   3/4
            12/16   3/4
            13/16   7/8
            14/16   7/8
            15/16   15/16
            16/16   1

        """
        good_denominator = _math.greatest_power_of_two_less_equal(self.denominator)
        current_numerator = self.numerator
        candidate = type(self)(current_numerator, good_denominator)
        while not candidate.is_assignable:
            current_numerator += 1
            candidate = type(self)(current_numerator, good_denominator)
        return candidate

    @property
    def equal_or_greater_power_of_two(self) -> "Duration":
        r"""
        Gets duration equal or just greater power of two.

        ..  container:: example

            Gets equal-or-greater power-of-two:

            >>> for numerator in range(1, 16 + 1):
            ...     duration = abjad.Duration(numerator, 16)
            ...     result = duration.equal_or_greater_power_of_two
            ...     sixteenths = duration.with_denominator(16)
            ...     print(f"{sixteenths!s}\t{result!s}")
            ...
            1/16    1/16
            2/16    1/8
            3/16    1/4
            4/16    1/4
            5/16    1/2
            6/16    1/2
            7/16    1/2
            8/16    1/2
            9/16    1
            10/16   1
            11/16   1
            12/16   1
            13/16   1
            14/16   1
            15/16   1
            16/16   1

        """
        denominator_exponent = -int(math.ceil(math.log(self, 2)))
        return type(self)(1, 2) ** denominator_exponent

    @property
    def equal_or_lesser_assignable(self) -> "Duration":
        r"""
        Gets assignable duration equal or just less than this duration.

        ..  container:: example

            Gets equal-or-lesser assignable duration:

            >>> for numerator in range(1, 16 + 1):
            ...     duration = abjad.Duration(numerator, 16)
            ...     result = duration.equal_or_lesser_assignable
            ...     sixteenths = duration.with_denominator(16)
            ...     print(f"{sixteenths!s}\t{result!s}")
            ...
            1/16    1/16
            2/16    1/8
            3/16    3/16
            4/16    1/4
            5/16    1/4
            6/16    3/8
            7/16    7/16
            8/16    1/2
            9/16    1/2
            10/16   1/2
            11/16   1/2
            12/16   3/4
            13/16   3/4
            14/16   7/8
            15/16   15/16
            16/16   1

        """
        good_denominator = self._least_power_of_two_greater_equal(self.denominator)
        current_numerator = self.numerator
        candidate = type(self)(current_numerator, good_denominator)
        while not candidate.is_assignable:
            current_numerator -= 1
            candidate = type(self)(current_numerator, good_denominator)
        return candidate

    @property
    def equal_or_lesser_power_of_two(self) -> "Duration":
        r"""
        Gets duration of the form ``d**2`` equal to or just less than this
        duration.

        ..  container:: example

            Gets equal-or-lesser power-of-two:

            >>> for numerator in range(1, 16 + 1):
            ...     duration = abjad.Duration(numerator, 16)
            ...     result = duration.equal_or_lesser_power_of_two
            ...     sixteenths = duration.with_denominator(16)
            ...     print(f"{sixteenths!s}\t{result!s}")
            ...
            1/16    1/16
            2/16    1/8
            3/16    1/8
            4/16    1/4
            5/16    1/4
            6/16    1/4
            7/16    1/4
            8/16    1/2
            9/16    1/2
            10/16   1/2
            11/16   1/2
            12/16   1/2
            13/16   1/2
            14/16   1/2
            15/16   1/2
            16/16   1

        """
        return type(self)(1, 2) ** self.exponent

    @property
    def exponent(self) -> int:
        r"""
        Gets base-2 exponent.

        ..  container:: example

            Gets equal-or-greater power-of-two:

            >>> for numerator in range(1, 16 + 1):
            ...     duration = abjad.Duration(numerator, 16)
            ...     exponent = duration.exponent
            ...     sixteenths = duration.with_denominator(16)
            ...     print(f"{sixteenths!s}    {duration.exponent!s}")
            ...
            1/16    4
            2/16    3
            3/16    3
            4/16    2
            5/16    2
            6/16    2
            7/16    2
            8/16    1
            9/16    1
            10/16    1
            11/16    1
            12/16    1
            13/16    1
            14/16    1
            15/16    1
            16/16    0

        """
        return -int(math.floor(math.log(self, 2)))

    @property
    def flag_count(self) -> int:
        r"""
        Gets flag count.

        ..  container:: example

            Gets flag count:

            >>> for n in range(1, 16 + 1):
            ...     duration = abjad.Duration(n, 64)
            ...     sixty_fourths = duration.with_denominator(64)
            ...     print(f"{sixty_fourths!s}\t{duration.flag_count}")
            ...
            1/64    4
            2/64    3
            3/64    3
            4/64    2
            5/64    2
            6/64    2
            7/64    2
            8/64    1
            9/64    1
            10/64   1
            11/64   1
            12/64   1
            13/64   1
            14/64   1
            15/64   1
            16/64   0

        Flag count defined equal to number of flags required to notate
        duration.
        """
        log = math.log(float(self.numerator) / self.denominator, 2)
        count = -int(math.floor(log)) - 2
        return max(count, 0)

    @property
    def is_dyadic_rational(self) -> bool:
        r"""
        Is true when duration is an integer power of two.

        ..  container:: example

            Is true when duration has power-of-two denominator:

            >>> for n in range(1, 16 + 1):
            ...     duration = abjad.Duration(1, n)
            ...     result = duration.is_dyadic_rational
            ...     print(f"{duration!s}\t{result}")
            ...
            1       True
            1/2     True
            1/3     False
            1/4     True
            1/5     False
            1/6     False
            1/7     False
            1/8     True
            1/9     False
            1/10    False
            1/11    False
            1/12    False
            1/13    False
            1/14    False
            1/15    False
            1/16    True

        """
        exponent = math.log(self.denominator, 2)
        return int(exponent) == exponent

    @property
    def implied_prolation(self):
        r"""
        Gets implied prolation.

        ..  container:: example

            Gets implied prolation:

            >>> for denominator in range(1, 16 + 1):
            ...     duration = abjad.Duration(1, denominator)
            ...     result = duration.implied_prolation
            ...     print(f"{duration!s}\t{result!s}")
            ...
            1       1
            1/2     1
            1/3     2/3
            1/4     1
            1/5     4/5
            1/6     2/3
            1/7     4/7
            1/8     1
            1/9     8/9
            1/10    4/5
            1/11    8/11
            1/12    2/3
            1/13    8/13
            1/14    4/7
            1/15    8/15
            1/16    1

        Returns multipler.
        """
        numerator = _math.greatest_power_of_two_less_equal(self.denominator)
        return Multiplier(numerator, self.denominator)

    @property
    def is_assignable(self) -> bool:
        r"""
        Is true when duration is assignable.

        ..  container:: example

            Is true when duration is assignable:

            >>> for numerator in range(0, 16 + 1):
            ...     duration = abjad.Duration(numerator, 16)
            ...     sixteenths = duration.with_denominator(16)
            ...     print(f"{sixteenths!s}\t{duration.is_assignable}")
            ...
            0/16    False
            1/16    True
            2/16    True
            3/16    True
            4/16    True
            5/16    False
            6/16    True
            7/16    True
            8/16    True
            9/16    False
            10/16   False
            11/16   False
            12/16   True
            13/16   False
            14/16   True
            15/16   True
            16/16   True

        """
        if 0 < self < 16:
            if _math.is_nonnegative_integer_power_of_two(self.denominator):
                if _math.is_assignable_integer(self.numerator):
                    return True
        return False

    @property
    def lilypond_duration_string(self) -> str:
        """
        Gets LilyPond duration string.

        ..  container:: example

            Gets LilyPond duration string:

                >>> abjad.Duration(3, 16).lilypond_duration_string
                '8.'

        Raises assignability error when duration is not assignable.
        """
        if not self.is_assignable:
            raise exceptions.AssignabilityError(self)
        undotted_rational = self.equal_or_lesser_power_of_two
        if undotted_rational <= 1:
            undotted_duration_string = str(undotted_rational.denominator)
        elif undotted_rational == type(self)(2, 1):
            undotted_duration_string = r"\breve"
        elif undotted_rational == type(self)(4, 1):
            undotted_duration_string = r"\longa"
        elif undotted_rational == type(self)(8, 1):
            undotted_duration_string = r"\maxima"
        else:
            raise ValueError(f"can not process undotted rational: {undotted_rational}")
        dot_count = self.dot_count
        dot_string = "." * dot_count
        dotted_duration_string = undotted_duration_string + dot_string
        return dotted_duration_string

    @property
    def pair(self) -> typing.Tuple[int, int]:
        """
        Gets numerator and denominator.

        ..  container:: example

            Gets pair:

            >>> abjad.Duration(3, 16).pair
            (3, 16)

        """
        return self.numerator, self.denominator

    @property
    def prolation_string(self) -> str:
        """
        Gets prolation string.

        ..  container:: example

            Gets prolation string:

            >>> abjad.Duration(3, 16).prolation_string
            '16:3'

        """
        return f"{self.denominator}:{self.numerator}"

    @property
    def reciprocal(self) -> "Duration":
        """
        Gets reciprocal.

        ..  container:: example

            Gets reciprocal:

            >>> abjad.Duration(3, 7).reciprocal
            Duration(7, 3)

        """
        return type(self)(self.denominator, self.numerator)

    ### PUBLIC FUNCTIONS ###

    @staticmethod
    def durations_to_nonreduced_fractions(
        durations: typing.List,
    ) -> typing.List["NonreducedFraction"]:
        """
        Changes ``durations`` to nonreduced fractions sharing least common
        denominator.

        ..  container:: example

            Changes durations to nonreduced fractions:

            >>> durations = [abjad.Duration(2, 4), 3, (5, 16)]
            >>> result = abjad.Duration.durations_to_nonreduced_fractions(durations)
            >>> for x in result:
            ...     x
            ...
            NonreducedFraction(8, 16)
            NonreducedFraction(48, 16)
            NonreducedFraction(5, 16)

        """
        durations_ = [Duration(_) for _ in durations]
        denominators = [_.denominator for _ in durations_]
        lcd = _math.least_common_multiple(*denominators)
        nonreduced_fractions = [
            NonreducedFraction(_).with_denominator(lcd) for _ in durations_
        ]
        return nonreduced_fractions

    @staticmethod
    def from_clock_string(clock_string) -> "Duration":
        """
        Initializes duration (in seconds) from clock string.

        ..  container:: example

            >>> abjad.Duration.from_clock_string("0'00''")
            Duration(0, 1)

            >>> abjad.Duration.from_clock_string("0'59''")
            Duration(59, 1)

            >>> abjad.Duration.from_clock_string("1'00''")
            Duration(60, 1)

            >>> abjad.Duration.from_clock_string("1'17''")
            Duration(77, 1)

        """
        minutes = 0
        if "'" in clock_string:
            tick_index = clock_string.find("'")
            minutes = clock_string[:tick_index]
            minutes = int(minutes)
        seconds = clock_string[-4:-2]
        seconds = int(seconds)
        seconds = 60 * minutes + seconds
        return Duration(seconds)

    @staticmethod
    def from_lilypond_duration_string(
        lilypond_duration_string,
    ) -> "Duration":
        """
        Initializes duration from LilyPond duration string.

        ..  container:: example

            Initializes duration from LilyPond duration string:

            >>> abjad.Duration.from_lilypond_duration_string('8.')
            Duration(3, 16)

        """
        fraction = Duration._initialize_from_lilypond_duration_string(
            lilypond_duration_string
        )
        return Duration(fraction)

    @staticmethod
    def is_token(argument) -> bool:
        """
        Is true when ``argument`` correctly initializes a duration.

        ..  container:: example

            Is true when expression is a duration token:

            >>> abjad.Duration.is_token('8.')
            True

        """
        try:
            Duration.__new__(Duration, argument)
            return True
        except Exception:
            return False

    def to_clock_string(self) -> str:
        r"""
        Changes duration to clock string.

        ..  container:: example

            Changes duration to clock string:

            >>> note = abjad.Note("c'4")
            >>> duration = abjad.Duration(117)
            >>> clock_string = duration.to_clock_string()
            >>> clock_string
            "1'57''"

            >>> string = f'"{clock_string}"'
            >>> markup = abjad.Markup(string, direction=abjad.Up)
            >>> abjad.attach(markup, note)
            >>> abjad.show(note) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(note)
                >>> print(string)
                c'4
                ^ \markup { 1'57'' }

        Rounds down to nearest second.
        """
        minutes = int(self / 60)
        seconds = str(int(self - minutes * 60)).zfill(2)
        clock_string = f"{minutes}'{seconds}''"
        return clock_string

    def with_denominator(self, denominator) -> "NonreducedFraction":
        """
        Changes duration to nonreduced fraction with ``denominator``.

        ..  container:: example

            Changes duration to nonreduced fraction:

            >>> duration = abjad.Duration(1, 4)
            >>> for denominator in (4, 8, 16, 32):
            ...     print(duration.with_denominator(denominator))
            ...
            1/4
            2/8
            4/16
            8/32

        """
        nonreduced_fraction = NonreducedFraction(self)
        return nonreduced_fraction.with_denominator(denominator)


class Multiplier(Duration):
    """
    Multiplier.

    ..  container:: example

        Initializes from integer numerator:

        >>> abjad.Multiplier(3)
        Multiplier(3, 1)

    ..  container:: example

        Initializes from integer numerator and denominator:

        >>> abjad.Multiplier(3, 16)
        Multiplier(3, 16)

    ..  container:: example

        Initializes from integer-equivalent numeric numerator:

        >>> abjad.Multiplier(3.0)
        Multiplier(3, 1)

    ..  container:: example

        Initializes from integer-equivalent numeric numerator and denominator:

        >>> abjad.Multiplier(3.0, 16)
        Multiplier(3, 16)

    ..  container:: example

        Initializes from integer-equivalent singleton:

        >>> abjad.Multiplier((3,))
        Multiplier(3, 1)

    ..  container:: example

        Initializes from integer-equivalent pair:

        >>> abjad.Multiplier((3, 16))
        Multiplier(3, 16)

    ..  container:: example

        Initializes from other duration:

        >>> abjad.Multiplier(abjad.Duration(3, 16))
        Multiplier(3, 16)

    ..  container:: example

        Intializes from fraction:

        >>> import quicktions
        >>> abjad.Multiplier(quicktions.Fraction(3, 16))
        Multiplier(3, 16)

    ..  container:: example

        Initializes from solidus string:

        >>> abjad.Multiplier('3/16')
        Multiplier(3, 16)

    ..  container:: example

        Initializes from nonreduced fraction:

        >>> abjad.Multiplier(abjad.NonreducedFraction(6, 32))
        Multiplier(3, 16)

    ..  container:: example

        Multipliers inherit from built-in fraction:

        >>> isinstance(abjad.Multiplier(3, 16), quicktions.Fraction)
        True

    ..  container:: example

        Multipliers are numbers:

        >>> import numbers

        >>> isinstance(abjad.Multiplier(3, 16), numbers.Number)
        True

    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### SPECIAL METHODS ###

    def __mul__(self, *arguments) -> "Duration":
        """
        Multiplier times duration gives duration.
        """
        if len(arguments) == 1 and type(arguments[0]) is Duration:
            return Duration(Duration.__mul__(self, *arguments))
        else:
            return Duration.__mul__(self, *arguments)

    ### PUBLIC PROPERTIES ###

    @staticmethod
    def from_dot_count(dot_count: int) -> "Multiplier":
        """
        Makes multiplier from ``dot_count``.

        ..  container:: example

            >>> abjad.Multiplier.from_dot_count(0)
            Multiplier(1, 1)

            >>> abjad.Multiplier.from_dot_count(1)
            Multiplier(3, 2)

            >>> abjad.Multiplier.from_dot_count(2)
            Multiplier(7, 4)

            >>> abjad.Multiplier.from_dot_count(3)
            Multiplier(15, 8)

            >>> abjad.Multiplier.from_dot_count(4)
            Multiplier(31, 16)

        """
        assert isinstance(dot_count, int), repr(dot_count)
        assert 0 <= dot_count, repr(dot_count)
        denominator = 2 ** dot_count
        numerator = 2 ** (dot_count + 1) - 1
        return Multiplier(numerator, denominator)

    def normalized(self):
        """
        Is true when mutliplier is greater than ``1/2`` and less than ``2``.

        ..  container:: example

            >>> abjad.Multiplier(3, 2).normalized()
            True

        Returns true or false.
        """
        return type(self)(1, 2) < self < type(self)(2)


class Offset(Duration):
    """
    Offset.

    ..  container:: example

        Initializes from integer numerator:

        >>> abjad.Offset(3)
        Offset((3, 1))

    ..  container:: example

        Initializes from integer numerator and denominator:

        >>> abjad.Offset(3, 16)
        Offset((3, 16))

    ..  container:: example

        Initializes from integer-equivalent numeric numerator:

        >>> abjad.Offset(3.0)
        Offset((3, 1))

    ..  container:: example

        Initializes from integer-equivalent numeric numerator and denominator:

        >>> abjad.Offset(3.0, 16)
        Offset((3, 16))

    ..  container:: example

        Initializes from integer-equivalent singleton:

        >>> abjad.Offset((3,))
        Offset((3, 1))

    ..  container:: example

        Initializes from integer-equivalent pair:

        >>> abjad.Offset((3, 16))
        Offset((3, 16))

    ..  container:: example

        Initializes from duration:

        >>> abjad.Offset(abjad.Duration(3, 16))
        Offset((3, 16))

    ..  container:: example

        Initializes from other offset:

        >>> abjad.Offset(abjad.Offset(3, 16))
        Offset((3, 16))

    ..  container:: example

        Initializes from other offset with displacement:

        >>> offset = abjad.Offset((3, 16), displacement=(-1, 16))
        >>> abjad.Offset(offset)
        Offset((3, 16), displacement=Duration(-1, 16))

    ..  container:: example

        Intializes from fraction:

        >>> import quicktions
        >>> abjad.Offset(quicktions.Fraction(3, 16))
        Offset((3, 16))

    ..  container:: example

        Initializes from solidus string:

        >>> abjad.Offset('3/16')
        Offset((3, 16))

    ..  container:: example

        Initializes from nonreduced fraction:

        >>> abjad.Offset(abjad.NonreducedFraction(6, 32))
        Offset((3, 16))

    ..  container:: example

        Offsets inherit from built-in fraction:

        >>> isinstance(abjad.Offset(3, 16), quicktions.Fraction)
        True

    ..  container:: example

        Offsets are numbers:

        >>> import numbers

        >>> isinstance(abjad.Offset(3, 16), numbers.Number)
        True

    ..  container:: example exception

        REGRESSION. Raises exception when new is attempted:

        >>> abjad.new(abjad.Offset((3, 16)))
        Traceback (most recent call last):
            ...
        Exception: low-level class not equipped for new():
            Offset((3, 16))

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_displacement",)

    ### CONSTRUCTOR ###

    def __new__(class_, *arguments, **keywords):
        displacement = None
        for argument in arguments:
            try:
                displacement = argument.displacement
                break
            except AttributeError:
                pass
        displacement = displacement or keywords.get("displacement")
        if displacement is not None:
            displacement = Duration(displacement)
        displacement = displacement or None
        if len(arguments) == 1 and isinstance(arguments[0], Duration):
            arguments = arguments[0].pair
        self = Duration.__new__(class_, *arguments)
        self._displacement = displacement
        return self

    ### SPECIAL METHODS ###

    def __copy__(self, *arguments) -> "Offset":
        """
        Copies offset.

        >>> import copy

        ..  container:: example

            Copies offset with displacement:

            >>> offset_1 = abjad.Offset((1, 4), displacement=(-1, 16))
            >>> offset_2 = copy.copy(offset_1)

            >>> offset_1
            Offset((1, 4), displacement=Duration(-1, 16))

            >>> offset_2
            Offset((1, 4), displacement=Duration(-1, 16))

            >>> offset_1 == offset_2
            True

            >>> offset_1 is offset_2
            False

        """
        return type(self)(self.pair, displacement=self.displacement)

    def __deepcopy__(self, *arguments) -> "Offset":
        """
        Deep copies offset.

        >>> import copy

        ..  container:: example

            Copies offset with displacement:

            >>> offset_1 = abjad.Offset((1, 4), displacement=(-1, 16))
            >>> offset_2 = copy.deepcopy(offset_1)

            >>> offset_1
            Offset((1, 4), displacement=Duration(-1, 16))

            >>> offset_2
            Offset((1, 4), displacement=Duration(-1, 16))

            >>> offset_1 == offset_2
            True

            >>> offset_1 is offset_2
            False

        """
        return self.__copy__(*arguments)

    def __eq__(self, argument) -> bool:
        """
        Is true when offset equals ``argument``.

        ..  container:: example

            With equal numerators, denominators and displacement:

            >>> offset_1 = abjad.Offset((1, 4), displacement=(-1, 16))
            >>> offset_2 = abjad.Offset((1, 4), displacement=(-1, 16))

            >>> offset_1 == offset_1
            True
            >>> offset_1 == offset_2
            True
            >>> offset_2 == offset_1
            True
            >>> offset_2 == offset_2
            True

        ..  container:: example

            With equal numerators and denominators but differing grace
            displacements:

            >>> offset_1 = abjad.Offset((1, 4), displacement=(-1, 8))
            >>> offset_2 = abjad.Offset((1, 4), displacement=(-1, 16))

            >>> offset_1 == offset_1
            True
            >>> offset_1 == offset_2
            False
            >>> offset_2 == offset_1
            False
            >>> offset_2 == offset_2
            True

        ..  container:: example

            With differing numerators and denominators. Ignores grace
            displacements:

            >>> offset_1 = abjad.Offset((1, 4))
            >>> offset_2 = abjad.Offset((1, 2), displacement=(-99))

            >>> offset_1 == offset_1
            True
            >>> offset_1 == offset_2
            False
            >>> offset_2 == offset_1
            False
            >>> offset_2 == offset_2
            True

        """
        if isinstance(argument, type(self)) and self.pair == argument.pair:
            return self._get_displacement() == argument._get_displacement()
        return super().__eq__(argument)

    def __ge__(self, argument) -> bool:
        """
        Is true when offset is greater than or equal to ``argument``.

        ..  container:: example

            With equal numerators, denominators and displacement:

            >>> offset_1 = abjad.Offset((1, 4), displacement=(-1, 16))
            >>> offset_2 = abjad.Offset((1, 4), displacement=(-1, 16))

            >>> offset_1 >= offset_1
            True
            >>> offset_1 >= offset_2
            True
            >>> offset_2 >= offset_1
            True
            >>> offset_2 >= offset_2
            True

        ..  container:: example

            With equal numerators and denominators but differing grace
            displacements:

            >>> offset_1 = abjad.Offset((1, 4), displacement=(-1, 8))
            >>> offset_2 = abjad.Offset((1, 4), displacement=(-1, 16))

            >>> offset_1 >= offset_1
            True
            >>> offset_1 >= offset_2
            False
            >>> offset_2 >= offset_1
            True
            >>> offset_2 >= offset_2
            True

        ..  container:: example

            With differing numerators and denominators. Ignores grace
            displacements:

            >>> offset_1 = abjad.Offset((1, 4))
            >>> offset_2 = abjad.Offset((1, 2), displacement=(-99))

            >>> offset_1 >= offset_1
            True
            >>> offset_1 >= offset_2
            False
            >>> offset_2 >= offset_1
            True
            >>> offset_2 >= offset_2
            True

        """
        if isinstance(argument, type(self)) and self.pair == argument.pair:
            return self._get_displacement() >= argument._get_displacement()
        return super().__ge__(argument)

    def __gt__(self, argument) -> bool:
        """
        Is true when offset is greater than ``argument``.

        ..  container:: example

            With equal numerators, denominators and displacement:

            >>> offset_1 = abjad.Offset((1, 4), displacement=(-1, 16))
            >>> offset_2 = abjad.Offset((1, 4), displacement=(-1, 16))

            >>> offset_1 > offset_1
            False
            >>> offset_1 > offset_2
            False
            >>> offset_2 > offset_1
            False
            >>> offset_2 > offset_2
            False

        ..  container:: example

            With equal numerators and denominators but differing grace
            displacements:

            >>> offset_1 = abjad.Offset((1, 4), displacement=(-1, 8))
            >>> offset_2 = abjad.Offset((1, 4), displacement=(-1, 16))

            >>> offset_1 > offset_1
            False
            >>> offset_1 > offset_2
            False
            >>> offset_2 > offset_1
            True
            >>> offset_2 > offset_2
            False

        ..  container:: example

            With differing numerators and denominators. Ignores grace
            displacements:

            >>> offset_1 = abjad.Offset((1, 4))
            >>> offset_2 = abjad.Offset((1, 2), displacement=(-99))

            >>> offset_1 > offset_1
            False
            >>> offset_1 > offset_2
            False
            >>> offset_2 > offset_1
            True
            >>> offset_2 > offset_2
            False

        """
        if isinstance(argument, type(self)) and self.pair == argument.pair:
            return self._get_displacement() > argument._get_displacement()
        return Duration.__gt__(self, argument)

    def __hash__(self) -> int:
        """
        Hashes offset.

        Redefined in tandem with __eq__.
        """
        return super().__hash__()

    def __le__(self, argument) -> bool:
        """
        Is true when offset is less than or equal to ``argument``.

        ..  container:: example

            With equal numerators, denominators and displacement:

            >>> offset_1 = abjad.Offset((1, 4), displacement=(-1, 16))
            >>> offset_2 = abjad.Offset((1, 4), displacement=(-1, 16))

            >>> offset_1 <= offset_1
            True
            >>> offset_1 <= offset_2
            True
            >>> offset_2 <= offset_1
            True
            >>> offset_2 <= offset_2
            True

        ..  container:: example

            With equal numerators and denominators but differing grace
            displacements:

            >>> offset_1 = abjad.Offset((1, 4), displacement=(-1, 8))
            >>> offset_2 = abjad.Offset((1, 4), displacement=(-1, 16))

            >>> offset_1 <= offset_1
            True
            >>> offset_1 <= offset_2
            True
            >>> offset_2 <= offset_1
            False
            >>> offset_2 <= offset_2
            True

        ..  container:: example

            With differing numerators and denominators. Ignores grace
            displacements:

            >>> offset_1 = abjad.Offset((1, 4))
            >>> offset_2 = abjad.Offset((1, 2), displacement=(-99))

            >>> offset_1 <= offset_1
            True
            >>> offset_1 <= offset_2
            True
            >>> offset_2 <= offset_1
            False
            >>> offset_2 <= offset_2
            True

        """
        if isinstance(argument, type(self)) and self.pair == argument.pair:
            return self._get_displacement() <= argument._get_displacement()
        return super().__le__(argument)

    def __lt__(self, argument) -> bool:
        """
        Is true when offset is less than ``argument``.

        ..  container:: example

            With equal numerators, denominators and displacement:

            >>> offset_1 = abjad.Offset((1, 4), displacement=(-1, 16))
            >>> offset_2 = abjad.Offset((1, 4), displacement=(-1, 16))

            >>> offset_1 < offset_1
            False
            >>> offset_1 < offset_2
            False
            >>> offset_2 < offset_1
            False
            >>> offset_2 < offset_2
            False

        ..  container:: example

            With equal numerators and denominators but differing nonzero grace
            displacements:

            >>> offset_1 = abjad.Offset((1, 4), displacement=(-1, 8))
            >>> offset_2 = abjad.Offset((1, 4), displacement=(-1, 16))

            >>> offset_1 < offset_1
            False
            >>> offset_1 < offset_2
            True
            >>> offset_2 < offset_1
            False
            >>> offset_2 < offset_2
            False

        ..  container:: example

            With equal numerators and denominators but differing zero-valued
            displacement:

            >>> offset_1 = abjad.Offset((1, 4), displacement=(-1, 8))
            >>> offset_2 = abjad.Offset((1, 4))

            >>> offset_1 < offset_1
            False
            >>> offset_1 < offset_2
            True
            >>> offset_2 < offset_1
            False
            >>> offset_2 < offset_2
            False

        ..  container:: example

            With differing numerators and denominators. Ignores grace
            displacements:

            >>> offset_1 = abjad.Offset((1, 4))
            >>> offset_2 = abjad.Offset((1, 2), displacement=(-99))

            >>> offset_1 < offset_1
            False
            >>> offset_1 < offset_2
            True
            >>> offset_2 < offset_1
            False
            >>> offset_2 < offset_2
            False

        """
        if isinstance(argument, type(self)) and self.pair == argument.pair:
            return self._get_displacement() < argument._get_displacement()
        return super().__lt__(argument)

    def __repr__(self) -> str:
        """
        Gets interpreter representation of offset.

        ..  container:: example

            >>> abjad.Offset(1, 4)
            Offset((1, 4))

            >>> abjad.Offset(1, 4, displacement=(-1, 16))
            Offset((1, 4), displacement=Duration(-1, 16))

        """
        return super().__repr__()

    def __sub__(self, argument):
        """
        Subtracts ``argument`` from offset.

        ..  container:: example

            Offset taken from offset returns duration:

            >>> abjad.Offset(2) - abjad.Offset(1, 2)
            Duration(3, 2)

            Duration taken from offset returns another offset:

            >>> abjad.Offset(2) - abjad.Duration(1, 2)
            Offset((3, 2))

        ..  container::

            Coerces ``argument`` to offset when ``argument`` is neither offset nor
            duration:

            >>> import quicktions
            >>> abjad.Offset(2) - quicktions.Fraction(1, 2)
            Duration(3, 2)

        """
        if isinstance(argument, type(self)):
            return Duration(super().__sub__(argument))
        elif isinstance(argument, Duration):
            return super().__sub__(argument)
        else:
            argument = type(self)(argument)
            return self - argument

    ### PRIVATE METHODS ###

    def _get_displacement(self):
        if self.displacement is None:
            return Duration(0)
        return self.displacement

    def _get_format_specification(self):
        names = []
        values = [(self.numerator, self.denominator)]
        if self._get_displacement():
            names = ["displacement"]
        return storage.FormatSpecification(
            client=self,
            storage_format_args_values=values,
            storage_format_is_indented=False,
            storage_format_keyword_names=names,
        )

    ### PUBLIC PROPERTIES ###

    @property
    def displacement(self) -> "Duration":
        """
        Gets displacement.

        ..  container:: example

            Gets displacement equal to none:

            >>> offset = abjad.Offset(1, 4)
            >>> offset.displacement is None
            True

        ..  container:: example

            Gets displacement equal to a negative sixteenth:

            >>> offset = abjad.Offset(1, 4, displacement=(-1, 16))
            >>> offset.displacement
            Duration(-1, 16)

        ..  container:: example

            Stores zero-valued displacement as none:

            >>> offset = abjad.Offset(1, 4, displacement=0)
            >>> offset.displacement is None
            True

            >>> offset
            Offset((1, 4))

        """
        return self._displacement


class NonreducedFraction(quicktions.Fraction):
    """
    Nonreduced fraction.

    ..  container:: example

        Initializes with an integer numerator and integer denominator:

        >>> abjad.NonreducedFraction(3, 6)
        NonreducedFraction(3, 6)

    ..  container:: example

        Initializes with only an integer denominator:

        >>> abjad.NonreducedFraction(3)
        NonreducedFraction(3, 1)

    ..  container:: example

        Initializes with an integer pair:

        >>> abjad.NonreducedFraction((3, 6))
        NonreducedFraction(3, 6)

    ..  container:: example

        Initializes with an integer singleton:

        >>> abjad.NonreducedFraction((3,))
        NonreducedFraction(3, 1)

    ..  container:: example

        Nonreduced fractions inherit from built-in fraction:

        >>> import quicktions
        >>> isinstance(abjad.NonreducedFraction(3, 6), quicktions.Fraction)
        True

    ..  container:: example

        Nonreduced fractions are numbers:

        >>> import numbers

        >>> isinstance(abjad.NonreducedFraction(3, 6), numbers.Number)
        True

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_numerator", "_denominator")

    ### CONSTRUCTOR ###

    # dummy docstring prevents Sphinx warnings
    def __new__(class_, *arguments):
        """
        Constructs nonreduced fraction.
        """
        is_fraction_like = False
        if len(arguments) == 1:
            try:
                numerator = arguments[0].numerator
                denominator = arguments[0].denominator
                is_fraction_like = True
            except AttributeError:
                pass
        if is_fraction_like:
            pass
        elif len(arguments) == 1 and isinstance(arguments[0], int):
            numerator = arguments[0]
            denominator = 1
        elif (
            len(arguments) == 1
            and isinstance(arguments[0], tuple)
            and len(arguments[0]) == 1
        ):
            numerator = arguments[0][0]
            denominator = 1
        elif (
            len(arguments) == 1
            and isinstance(arguments[0], tuple)
            and isinstance(arguments[0][0], int)
            and isinstance(arguments[0][1], int)
        ):
            numerator, denominator = arguments[0]
        elif len(arguments) == 1 and isinstance(arguments[0], str):
            numerator, denominator = class_._parse_input_string(arguments[0])
        elif (
            isinstance(arguments, tuple)
            and len(arguments) == 2
            and isinstance(arguments[0], int)
            and isinstance(arguments[1], int)
        ):
            numerator = arguments[0]
            denominator = arguments[1]
        elif len(arguments) == 0:
            numerator = 0
            denominator = 1
        else:
            raise ValueError(f"can not initialize {class_.__name__}: {arguments!r}.")
        numerator *= _math.sign(denominator)
        denominator = abs(denominator)
        self = quicktions.Fraction.__new__(class_, numerator, denominator)
        self._numerator = numerator
        self._denominator = denominator
        return self

    ### INITIALIZER ###

    def __init__(self, *arguments):
        """
        Dummy initializer to satisfy mypy.
        """
        pass

    ### SPECIAL METHODS ###

    def __abs__(self) -> "NonreducedFraction":
        """
        Gets absolute value of nonreduced fraction.

        ..  container:: example

            >>> abs(abjad.NonreducedFraction(-3, 3))
            NonreducedFraction(3, 3)

        """
        pair = (abs(self.numerator), self.denominator)
        return type(self)(pair)

    def __add__(self, argument) -> "NonreducedFraction":
        """
        Adds ``argument`` to nonreduced fraction.

        ..  container:: example

            >>> abjad.NonreducedFraction(3, 3) + 1
            NonreducedFraction(6, 3)

            >>> 1 + abjad.NonreducedFraction(3, 3)
            NonreducedFraction(6, 3)

        """
        if isinstance(argument, int):
            numerator = self.numerator + argument * self.denominator
            pair = (numerator, self.denominator)
            return type(self)(pair)
        if getattr(argument, "denominator", "foo") == "foo":
            raise ValueError(argument)
        if self.denominator == argument.denominator:
            numerator = self.numerator + argument.numerator
            pair = (numerator, self.denominator)
            return type(self)(pair)
        else:
            denominators = [self.denominator, argument.denominator]
            denominator = _math.least_common_multiple(*denominators)
            self_multiplier = denominator // self.denominator
            argument_multiplier = denominator // argument.denominator
            self_numerator = self_multiplier * self.numerator
            argument_numerator = argument_multiplier * argument.numerator
            pair = (self_numerator + argument_numerator, denominator)
            return type(self)(pair)

    def __copy__(self, *arguments) -> "NonreducedFraction":
        """
        Copies nonreduced fraction.

        >>> import copy

        ..  container:: example

            >>> fraction_1 = abjad.NonreducedFraction(4, 6)
            >>> fraction_2 = copy.copy(fraction_1)

            >>> fraction_1
            NonreducedFraction(4, 6)

            >>> fraction_2
            NonreducedFraction(4, 6)

            >>> fraction_1 == fraction_2
            True

            >>> fraction_1 is fraction_2
            False

        """
        return type(self)(self.pair)

    def __deepcopy__(self, *arguments) -> "NonreducedFraction":
        """
        Deep copies nonreduced fraction.

        >>> import copy

        ..  container:: example

            >>> fraction_1 = abjad.NonreducedFraction(4, 6)
            >>> fraction_2 = copy.deepcopy(fraction_1)

            >>> fraction_1
            NonreducedFraction(4, 6)

            >>> fraction_2
            NonreducedFraction(4, 6)

            >>> fraction_1 == fraction_2
            True

            >>> fraction_1 is fraction_2
            False

        """
        return self.__copy__(*arguments)

    def __div__(self, argument) -> "NonreducedFraction":
        """
        Divides nonreduced fraction by ``argument``.

        ..  container:: example

            >>> abjad.NonreducedFraction(3, 3) / 1
            NonreducedFraction(3, 3)

        """
        denominators = [self.denominator]
        if isinstance(argument, type(self)):
            denominators.append(argument.denominator)
            argument = argument.reduce()
        fraction = self.reduce() / argument
        return self._fraction_with_denominator(fraction, max(denominators))

    def __eq__(self, argument) -> bool:
        """
        Is true when ``argument`` equals nonreduced fraction.

        ..  container:: example

            >>> abjad.NonreducedFraction(3, 3) == 1
            True

        """
        return self.reduce() == argument

    def __ge__(self, argument) -> bool:
        """
        Is true when nonreduced fraction is greater than or equal to
        ``argument``.

        ..  container:: example

            >>> abjad.NonreducedFraction(3, 3) >= 1
            True

        """
        return self.reduce() >= argument

    def __gt__(self, argument) -> bool:
        """
        Is true when nonreduced fraction is greater than ``argument``.

        ..  container:: example

            >>> abjad.NonreducedFraction(3, 3) > 1
            False

        """
        return self.reduce() > argument

    def __hash__(self) -> int:
        """
        Hashes nonreduced fraction.

        Required to be explicitly redefined on Python 3 if __eq__ changes.
        """
        return super().__hash__()

    def __le__(self, argument) -> bool:
        """
        Is true when nonreduced fraction is less than or equal to ``argument``.

        ..  container:: example

            >>> abjad.NonreducedFraction(3, 3) <= 1
            True

        """
        return self.reduce() <= argument

    def __lt__(self, argument) -> bool:
        """
        Is true when nonreduced fraction is less than ``argument``.

        ..  container:: example

            >>> abjad.NonreducedFraction(3, 3) < 1
            False

        Returns true or false.
        """
        return self.reduce() < argument

    def __mul__(self, argument) -> "NonreducedFraction":
        """
        Multiplies nonreduced fraction by ``argument``.

        ..  container:: example

            >>> abjad.NonreducedFraction(3, 3) * 3
            NonreducedFraction(9, 3)

        """
        denominators = [self.denominator]
        if isinstance(argument, type(self)):
            denominators.append(argument.denominator)
            argument = argument.reduce()
        fraction = self.reduce() * argument
        return self._fraction_with_denominator(fraction, max(denominators))

    def __neg__(self) -> "NonreducedFraction":
        """
        Negates nonreduced fraction.

        ..  container:: example

            >>> -abjad.NonreducedFraction(3, 3)
            NonreducedFraction(-3, 3)

        """
        pair = (-self.numerator, self.denominator)
        return type(self)(pair)

    def __pow__(self, argument) -> "NonreducedFraction":
        """
        Raises nonreduced fraction to ``argument``.

        ..  container:: example

            >>> abjad.NonreducedFraction(3, 6) ** -1
            NonreducedFraction(6, 3)

        """
        if argument == -1:
            pair = (self.denominator, self.numerator)
            return type(self)(pair)
        return super().__pow__(argument)

    def __radd__(self, argument) -> "NonreducedFraction":
        """
        Adds nonreduced fraction to ``argument``.

        ..  container:: example

            >>> 1 + abjad.NonreducedFraction(3, 3)
            NonreducedFraction(6, 3)

        """
        return self + argument

    def __rdiv__(self, argument) -> "NonreducedFraction":
        """
        Divides ``argument`` by nonreduced fraction.

        ..  container:: example

            >>> 1 / abjad.NonreducedFraction(3, 3)
            NonreducedFraction(3, 3)

        """
        denominators = [self.denominator]
        if isinstance(argument, type(self)):
            denominators.append(argument.denominator)
            argument = argument.reduce()
        fraction = argument / self.reduce()
        return self._fraction_with_denominator(fraction, max(denominators))

    __rtruediv__ = __rdiv__

    def __repr__(self) -> str:
        """
        Gets interpreter representation of nonreduced fraction.

        ..  container:: example

            >>> abjad.NonreducedFraction(3, 6)
            NonreducedFraction(3, 6)

        """
        return storage.StorageFormatManager(self).get_repr_format()

    def __rmul__(self, argument) -> "NonreducedFraction":
        """
        Multiplies ``argument`` by nonreduced fraction.

        ..  container:: example

            >>> 3 * abjad.NonreducedFraction(3, 3)
            NonreducedFraction(9, 3)

        """
        return self * argument

    def __rsub__(self, argument) -> "NonreducedFraction":
        """
        Subtracts nonreduced fraction from ``argument``.

        ..  container:: example

            >>> 1 - abjad.NonreducedFraction(3, 3)
            NonreducedFraction(0, 3)

        """
        return -self + argument

    def __str__(self) -> str:
        """
        Gets string representation of nonreduced fraction.

        ..  container:: example

            >>> fraction = abjad.NonreducedFraction(-6, 3)

            >>> str(fraction)
            '-6/3'

        """
        return f"{self.numerator}/{self.denominator}"

    def __sub__(self, argument) -> "NonreducedFraction":
        """
        Subtracts ``argument`` from nonreduced fraction.

        ..  container:: example

            >>> abjad.NonreducedFraction(3, 3) - 2
            NonreducedFraction(-3, 3)

            >>> abjad.NonreducedFraction(5, 4) - abjad.NonreducedFraction(18, 16)
            NonreducedFraction(2, 16)

            >>> abjad.NonreducedFraction(18, 16) - abjad.NonreducedFraction(5, 4)
            NonreducedFraction(-2, 16)

        """
        denominators = [self.denominator]
        if isinstance(argument, type(self)):
            denominators.append(argument.denominator)
            argument = argument.reduce()
        fraction = self.reduce() - argument
        return self._fraction_with_denominator(fraction, max(denominators))

    def __truediv__(self, argument) -> "NonreducedFraction":
        """
        Divides nonreduced fraction in Python 3.
        """
        return self.__div__(argument)

    ### PRIVATE METHODS ###

    def _fraction_with_denominator(self, fraction, denominator):
        denominators = [denominator, fraction.denominator]
        denominator = _math.least_common_multiple(*denominators)
        result = type(self)(fraction)
        result = result.with_denominator(denominator)
        return result

    def _get_format_specification(self):
        return storage.FormatSpecification(
            client=self,
            repr_is_indented=False,
            storage_format_args_values=[self.numerator, self.denominator],
            storage_format_is_indented=False,
            storage_format_keyword_names=[],
        )

    @staticmethod
    def _parse_input_string(string):
        if "/" in string:
            numerator, denominator = string.split("/")
            numerator = int(numerator)
            denominator = int(denominator)
        else:
            numerator = int(string)
            denominator = 1
        return numerator, denominator

    ### PUBLIC METHODS ###

    @staticmethod
    def durations_to_nonreduced_fractions(
        durations: typing.List,
    ) -> typing.List["NonreducedFraction"]:
        """
        Changes ``durations`` to nonreduced fractions sharing least common
        denominator.

        ..  container:: example

            Changes durations to nonreduced fractions:

            >>> durations = [abjad.Duration(2, 4), 3, (5, 16)]
            >>> class_ = abjad.NonreducedFraction
            >>> result = class_.durations_to_nonreduced_fractions(durations)
            >>> for x in result:
            ...     x
            ...
            NonreducedFraction(8, 16)
            NonreducedFraction(48, 16)
            NonreducedFraction(5, 16)

        """
        durations_ = [Duration(_) for _ in durations]
        denominators = [_.denominator for _ in durations_]
        lcd = _math.least_common_multiple(*denominators)
        nonreduced_fractions = [
            NonreducedFraction(_).with_denominator(lcd) for _ in durations_
        ]
        return nonreduced_fractions

    def multiply(self, multiplier, preserve_numerator=False) -> "NonreducedFraction":
        """
        Multiplies nonreduced fraction by ``multiplier`` with numerator
        preservation where possible.

        ..  container:: example

            >>> fraction = abjad.NonreducedFraction(9, 16)

            >>> fraction.multiply((2, 3), preserve_numerator=True)
            NonreducedFraction(9, 24)

            >>> fraction.multiply((1, 2), preserve_numerator=True)
            NonreducedFraction(9, 32)

            >>> fraction.multiply((5, 6), preserve_numerator=True)
            NonreducedFraction(45, 96)

            >>> fraction = abjad.NonreducedFraction(3, 8)

            >>> fraction.multiply((2, 3), preserve_numerator=True)
            NonreducedFraction(3, 12)

        """
        if preserve_numerator:
            multiplier = quicktions.Fraction(*multiplier)
            self_denominator = self.denominator
            candidate_result_denominator = self_denominator / multiplier
            if candidate_result_denominator.denominator == 1:
                pair = (self.numerator, candidate_result_denominator.numerator)
                return type(self)(pair)
            else:
                denominator = candidate_result_denominator.denominator
                result_numerator = self.numerator * denominator
                result_denominator = candidate_result_denominator.numerator
                pair = (result_numerator, result_denominator)
                return type(self)(pair)
        else:
            return multiplier * self

    def multiply_with_cross_cancelation(self, multiplier) -> "NonreducedFraction":
        """
        Multiplies nonreduced fraction by ``argument`` with cross-cancelation.

        ..  container:: example

            >>> fraction = abjad.NonreducedFraction(4, 8)

            >>> fraction.multiply_with_cross_cancelation((2, 3))
            NonreducedFraction(4, 12)

            >>> fraction.multiply_with_cross_cancelation((4, 1))
            NonreducedFraction(4, 2)

            >>> fraction.multiply_with_cross_cancelation((3, 5))
            NonreducedFraction(12, 40)

            >>> fraction.multiply_with_cross_cancelation((6, 5))
            NonreducedFraction(12, 20)

            >>> fraction = abjad.NonreducedFraction(5, 6)
            >>> fraction.multiply_with_cross_cancelation((6, 5))
            NonreducedFraction(1, 1)

        """
        multiplier = quicktions.Fraction(*multiplier)
        self_numerator_factors = _math.factors(self.numerator)
        multiplier_denominator_factors = _math.factors(multiplier.denominator)
        for factor in multiplier_denominator_factors[:]:
            if factor in self_numerator_factors:
                self_numerator_factors.remove(factor)
                multiplier_denominator_factors.remove(factor)
        self_denominator_factors = _math.factors(self.denominator)
        multiplier_numerator_factors = _math.factors(multiplier.numerator)
        for factor in multiplier_numerator_factors[:]:
            if factor in self_denominator_factors:
                self_denominator_factors.remove(factor)
                multiplier_numerator_factors.remove(factor)
        result_numerator_factors = self_numerator_factors + multiplier_numerator_factors
        result_denominator_factors = (
            self_denominator_factors + multiplier_denominator_factors
        )
        result_numerator = 1
        for factor in result_numerator_factors:
            result_numerator *= factor
        result_denominator = 1
        for factor in result_denominator_factors:
            result_denominator *= factor
        pair = (result_numerator, result_denominator)
        return type(self)(pair)

    def multiply_without_reducing(self, argument) -> "NonreducedFraction":
        """
        Multiplies nonreduced fraction by ``argument`` without reducing.

        ..  container:: example

            >>> fraction = abjad.NonreducedFraction(3, 8)

            >>> fraction.multiply_without_reducing((3, 3))
            NonreducedFraction(9, 24)

            >>> fraction = abjad.NonreducedFraction(4, 8)

            >>> fraction.multiply_without_reducing((4, 5))
            NonreducedFraction(16, 40)

            >>> fraction.multiply_without_reducing((3, 4))
            NonreducedFraction(12, 32)

        """
        argument = type(self)(argument)
        numerator = self.numerator * argument.numerator
        denominator = self.denominator * argument.denominator
        pair = (numerator, denominator)
        return type(self)(pair)

    def reduce(self) -> quicktions.Fraction:
        """
        Reduces nonreduced fraction.

        ..  container:: example

            >>> fraction = abjad.NonreducedFraction(-6, 3)

            >>> fraction.reduce()
            Fraction(-2, 1)

        """
        return quicktions.Fraction(self.numerator, self.denominator)

    def with_denominator(self, denominator) -> "NonreducedFraction":
        """
        Returns new nonreduced fraction with integer ``denominator``.

        ..  container:: example

            >>> abjad.NonreducedFraction(3, 6).with_denominator(12)
            NonreducedFraction(6, 12)

        ..  container:: example

            >>> for numerator in range(12):
            ...     fraction = abjad.NonreducedFraction(numerator, 6)
            ...     print(fraction, fraction.with_denominator(12))
            ...
            0/6 0/12
            1/6 2/12
            2/6 4/12
            3/6 6/12
            4/6 8/12
            5/6 10/12
            6/6 12/12
            7/6 14/12
            8/6 16/12
            9/6 18/12
            10/6 20/12
            11/6 22/12

            >>> for numerator in range(12):
            ...     fraction = abjad.NonreducedFraction(numerator, 6)
            ...     print(fraction, fraction.with_denominator(8))
            ...
            0/6 0/8
            1/6 1/6
            2/6 2/6
            3/6 4/8
            4/6 4/6
            5/6 5/6
            6/6 8/8
            7/6 7/6
            8/6 8/6
            9/6 12/8
            10/6 10/6
            11/6 11/6

            >>> for numerator in range(12):
            ...     fraction = abjad.NonreducedFraction(numerator, 6)
            ...     print(fraction, fraction.with_denominator(12))
            ...
            0/6 0/12
            1/6 2/12
            2/6 4/12
            3/6 6/12
            4/6 8/12
            5/6 10/12
            6/6 12/12
            7/6 14/12
            8/6 16/12
            9/6 18/12
            10/6 20/12
            11/6 22/12

        """
        current_numerator, current_denominator = self.pair
        multiplier = quicktions.Fraction(denominator, current_denominator)
        new_numerator = multiplier * current_numerator
        new_denominator = multiplier * current_denominator
        if new_numerator.denominator == 1 and new_denominator.denominator == 1:
            pair = (new_numerator.numerator, new_denominator.numerator)
        else:
            pair = (current_numerator, current_denominator)
        return type(self)(pair)

    def with_multiple_of_denominator(self, denominator) -> "NonreducedFraction":
        """
        Returns new nonreduced fraction with multiple of integer
        ``denominator``.

        ..  container:: example

            >>> fraction = abjad.NonreducedFraction(3, 6)

            >>> fraction.with_multiple_of_denominator(5)
            NonreducedFraction(5, 10)

        ..  container:: example

            >>> abjad.NonreducedFraction(1, 2).with_multiple_of_denominator(2)
            NonreducedFraction(1, 2)

            >>> abjad.NonreducedFraction(1, 2).with_multiple_of_denominator(4)
            NonreducedFraction(2, 4)

            >>> abjad.NonreducedFraction(1, 2).with_multiple_of_denominator(8)
            NonreducedFraction(4, 8)

            >>> abjad.NonreducedFraction(1, 2).with_multiple_of_denominator(16)
            NonreducedFraction(8, 16)

        """
        result = self.with_denominator(denominator)
        while not result.denominator == denominator:
            denominator *= 2
            result = result.with_denominator(denominator)
        return result

    ### PUBLIC PROPERTIES ###

    @property
    def denominator(self) -> int:
        """
        Gets denominator of nonreduced fraction.

        ..  container:: example

            >>> abjad.NonreducedFraction(-6, 3).denominator
            3

        """
        return self._denominator

    @property
    def imag(self) -> int:
        """
        Gets zero because nonreduced fractions have no imaginary part.

        ..  container:: example

            >>> abjad.NonreducedFraction(-6, 3).imag
            0

        """
        return 0

    @property
    def numerator(self) -> int:
        """
        Gets numerator of nonreduced fraction.

        ..  container:: example

            >>> abjad.NonreducedFraction(-6, 3).numerator
            -6

        """
        return self._numerator

    @property
    def pair(self) -> typing.Tuple[int, int]:
        """
        Gets (numerator, denominator) pair of nonreduced fraction.

        ..  container:: example

            >>> abjad.NonreducedFraction(-6, 3).pair
            (-6, 3)

        """
        return self.numerator, self.denominator

    @property
    def real(self) -> "NonreducedFraction":
        """
        Gets nonreduced fraction because nonreduced fractions are their own
        real component.

        ..  container:: example

            >>> abjad.NonreducedFraction(-6, 3).real
            NonreducedFraction(-6, 3)

        """
        return self



import typing

from ..bundle import LilyPondFormatBundle
from ..overrides import TweakInterface
from ..storage import StorageFormatManager


class StartPianoPedal:
    r"""
    LilyPond ``\sustainOn``, ``\sostenutoOn``, ``\unaCorda`` commands.

    ..  container:: example

        >>> staff = abjad.Staff("c'4 d' e' r")
        >>> start_piano_pedal = abjad.StartPianoPedal()
        >>> abjad.tweak(start_piano_pedal).color = "#blue"
        >>> abjad.tweak(start_piano_pedal).parent_alignment_X = abjad.Center
        >>> abjad.attach(start_piano_pedal, staff[0])
        >>> stop_piano_pedal = abjad.StopPianoPedal()
        >>> abjad.attach(stop_piano_pedal, staff[1])

        >>> start_piano_pedal = abjad.StartPianoPedal()
        >>> abjad.tweak(start_piano_pedal).color = "#red"
        >>> abjad.attach(start_piano_pedal, staff[1])
        >>> stop_piano_pedal = abjad.StopPianoPedal()
        >>> abjad.attach(stop_piano_pedal, staff[2])

        >>> start_piano_pedal = abjad.StartPianoPedal()
        >>> abjad.tweak(start_piano_pedal).color = "#green"
        >>> abjad.attach(start_piano_pedal, staff[2])
        >>> stop_piano_pedal = abjad.StopPianoPedal()
        >>> abjad.attach(stop_piano_pedal, staff[3])

        >>> abjad.override(staff).SustainPedalLineSpanner.staff_padding = 5
        >>> abjad.setting(staff).pedalSustainStyle = "#'mixed"
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            \with
            {
                \override SustainPedalLineSpanner.staff-padding = 5
                pedalSustainStyle = #'mixed
            }
            {
                c'4
                - \tweak color #blue
                - \tweak parent-alignment-X #center
                \sustainOn
                d'4
                \sustainOff
                - \tweak color #red
                \sustainOn
                e'4
                \sustainOff
                - \tweak color #green
                \sustainOn
                r4
                \sustainOff
            }

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_kind", "_tweaks")

    _context = "StaffGroup"

    _persistent = True

    _parameter = "PEDAL"

    ### INITIALIZER ###

    def __init__(self, kind: str = None, *, tweaks: TweakInterface = None) -> None:
        if kind is not None:
            assert kind in ("sustain", "sostenuto", "corda")
        self._kind = kind
        if tweaks is not None:
            assert isinstance(tweaks, TweakInterface), repr(tweaks)
        self._tweaks = TweakInterface.set_tweaks(self, tweaks)

    ### SPECIAL METHODS ###

    def __eq__(self, argument) -> bool:
        """
        Is true when all initialization values of Abjad value object equal
        the initialization values of ``argument``.
        """
        return StorageFormatManager.compare_objects(self, argument)

    def __hash__(self) -> int:
        """
        Hashes Abjad value object.
        """
        hash_values = StorageFormatManager(self).get_hash_values()
        try:
            result = hash(hash_values)
        except TypeError:
            raise TypeError(f"unhashable type: {self}")
        return result

    def __repr__(self) -> str:
        """
        Gets interpreter representation.
        """
        return StorageFormatManager(self).get_repr_format()

    ### PRIVATE METHODS ###

    def _get_lilypond_format_bundle(self, component=None):
        bundle = LilyPondFormatBundle()
        if self.tweaks:
            tweaks = self.tweaks._list_format_contributions()
            bundle.after.spanner_starts.extend(tweaks)
        if self.kind == "corda":
            string = r"\unaCorda"
        elif self.kind == "sostenuto":
            string = r"\sostenutoOn"
        else:
            assert self.kind in ("sustain", None)
            string = r"\sustainOn"
        bundle.after.spanner_starts.append(string)
        return bundle

    ### PUBLIC PROPERTIES ###

    @property
    def context(self) -> str:
        """
        Returns (historically conventional) context ``'StaffGroup'``.

        ..  container:: example

            >>> abjad.StartPianoPedal().context
            'StaffGroup'

        Class constant.

        Override with ``abjad.attach(..., context='...')``.
        """
        return self._context

    @property
    def kind(self) -> typing.Optional[str]:
        """
        Gets kind.
        """
        return self._kind

    @property
    def parameter(self) -> str:
        """
        Returns ``'PEDAL'``.

        ..  container:: example

            >>> abjad.StartPianoPedal().parameter
            'PEDAL'

        Class constant.
        """
        return self._parameter

    @property
    def persistent(self) -> bool:
        """
        Is true.

        ..  container:: example

            >>> abjad.StartPianoPedal().persistent
            True

        Class constant.
        """
        return self._persistent

    @property
    def spanner_start(self) -> bool:
        """
        Is true.

        ..  container:: example

            >>> abjad.StartPianoPedal().spanner_start
            True

        """
        return True

    @property
    def tweaks(self) -> typing.Optional[TweakInterface]:
        r"""
        Gets tweaks

        ..  container:: example

            REGRESSION. Tweaks survive copy:

            >>> import copy
            >>> start_piano_pedal = abjad.StartPianoPedal()
            >>> abjad.tweak(start_piano_pedal).color = "#blue"
            >>> string = abjad.storage(start_piano_pedal)
            >>> print(string)
            abjad.StartPianoPedal(
                tweaks=TweakInterface(('_literal', None), ('color', '#blue')),
                )

            >>> start_piano_pedal_2 = copy.copy(start_piano_pedal)
            >>> string = abjad.storage(start_piano_pedal_2)
            >>> print(string)
            abjad.StartPianoPedal(
                tweaks=TweakInterface(('_literal', None), ('color', '#blue')),
                )

        """
        return self._tweaks



import typing

from ..bundle import LilyPondFormatBundle
from ..overrides import TweakInterface
from ..storage import StorageFormatManager


class Glissando:
    r"""
    LilyPond ``\glissando`` command.

    ..  container:: example

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> glissando = abjad.Glissando()
        >>> abjad.tweak(glissando).color = "#blue"
        >>> abjad.attach(glissando, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'4
                - \tweak color #blue
                \glissando
                d'4
                e'4
                f'4
            }

    ..  container:: example

        >>> abjad.Glissando()
        Glissando()

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_allow_repeats",
        "_allow_ties",
        "_parenthesize_repeats",
        "_stems",
        "_style",
        "_zero_padding",
        "_tweaks",
    )

    _context = "Voice"

    _persistent = True

    ### INITIALIZER ###

    def __init__(
        self,
        *,
        allow_repeats: bool = None,
        allow_ties: bool = None,
        parenthesize_repeats: bool = None,
        stems: bool = None,
        style: str = None,
        tweaks: TweakInterface = None,
        zero_padding: bool = None,
    ) -> None:
        if allow_repeats is not None:
            allow_repeats = bool(allow_repeats)
        self._allow_repeats = allow_repeats
        if allow_ties is not None:
            allow_ties = bool(allow_ties)
        self._allow_ties = allow_ties
        if parenthesize_repeats is not None:
            parenthesize_repeats = bool(parenthesize_repeats)
        self._parenthesize_repeats = parenthesize_repeats
        if stems is not None:
            stems = bool(stems)
        self._stems = stems
        if style is not None:
            assert isinstance(style, str), repr(style)
        self._style = style
        if tweaks is not None:
            assert isinstance(tweaks, TweakInterface), repr(tweaks)
        self._tweaks = TweakInterface.set_tweaks(self, tweaks)
        if zero_padding is not None:
            zero_padding = bool(zero_padding)
        self._zero_padding = zero_padding

    ### SPECIAL METHODS ###

    def __eq__(self, argument) -> bool:
        """
        Is true when all initialization values of Abjad value object equal
        the initialization values of ``argument``.
        """
        return StorageFormatManager.compare_objects(self, argument)

    def __hash__(self) -> int:
        """
        Hashes Abjad value object.
        """
        hash_values = StorageFormatManager(self).get_hash_values()
        try:
            result = hash(hash_values)
        except TypeError:
            raise TypeError(f"unhashable type: {self}")
        return result

    def __repr__(self) -> str:
        """
        Gets interpreter representation.
        """
        return StorageFormatManager(self).get_repr_format()

    ### PRIVATE METHODS ###

    def _get_lilypond_format_bundle(self, component=None):
        bundle = LilyPondFormatBundle()
        strings = []
        if self.zero_padding:
            strings.append(r"- \abjad-zero-padding-glissando")
        if self.tweaks:
            tweaks = self.tweaks._list_format_contributions()
            strings.extend(tweaks)
        strings.append(r"\glissando")
        bundle.after.spanner_starts.extend(strings)
        return bundle

    ### PUBLIC PROPERTIES ###

    @property
    def context(self) -> str:
        """
        Returns (historically conventional) context ``'Voice'``.

        ..  container:: example

            >>> abjad.Glissando().context
            'Voice'

        Class constant.

        Override with ``abjad.attach(..., context='...')``.
        """
        return self._context

    @property
    def persistent(self) -> bool:
        """
        Is true.

        ..  container:: example

            >>> abjad.Glissando().persistent
            True

        Class constant.
        """
        return self._persistent

    @property
    def tweaks(self) -> typing.Optional[TweakInterface]:
        r"""
        Gets tweaks

        ..  container:: example

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> glissando = abjad.Glissando()
            >>> abjad.tweak(glissando).color = "#blue"
            >>> abjad.attach(glissando, staff[0])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(staff)
                >>> print(string)
                \new Staff
                {
                    c'4
                    - \tweak color #blue
                    \glissando
                    d'4
                    e'4
                    f'4
                }

        """
        return self._tweaks

    @property
    def zero_padding(self) -> typing.Optional[bool]:
        r"""
        Is true when glissando formats with zero padding and no extra space for
        dots.

        ..  container:: example

            >>> staff = abjad.Staff("d'8 d'4. d'4. d'8")
            >>> abjad.glissando(
            ...     staff[:],
            ...     allow_repeats=True,
            ...     zero_padding=True,
            ...     )
            >>> for note in staff[1:]:
            ...     abjad.override(note).NoteHead.transparent = True
            ...     abjad.override(note).NoteHead.X_extent = "#'(0 . 0)"
            ...
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(staff)
                >>> print(string)
                \new Staff
                {
                    d'8
                    - \abjad-zero-padding-glissando
                    \glissando
                    \once \override NoteHead.X-extent = #'(0 . 0)
                    \once \override NoteHead.transparent = ##t
                    d'4.
                    - \abjad-zero-padding-glissando
                    \glissando
                    \once \override NoteHead.X-extent = #'(0 . 0)
                    \once \override NoteHead.transparent = ##t
                    d'4.
                    - \abjad-zero-padding-glissando
                    \glissando
                    \once \override NoteHead.X-extent = #'(0 . 0)
                    \once \override NoteHead.transparent = ##t
                    d'8
                }

        ..  container:: example

            >>> staff = abjad.Staff("c'8. d'8. e'8. f'8.")
            >>> abjad.glissando(
            ...     staff[:],
            ...     zero_padding=True,
            ...     )
            >>> for note in staff[1:-1]:
            ...     abjad.override(note).NoteHead.transparent = True
            ...     abjad.override(note).NoteHead.X_extent = "#'(0 . 0)"
            ...
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(staff)
                >>> print(string)
                \new Staff
                {
                    c'8.
                    - \abjad-zero-padding-glissando
                    \glissando
                    \once \override NoteHead.X-extent = #'(0 . 0)
                    \once \override NoteHead.transparent = ##t
                    d'8.
                    - \abjad-zero-padding-glissando
                    \glissando
                    \once \override NoteHead.X-extent = #'(0 . 0)
                    \once \override NoteHead.transparent = ##t
                    e'8.
                    - \abjad-zero-padding-glissando
                    \glissando
                    f'8.
                }

        """
        return self._zero_padding



import collections
import typing

import quicktions

from . import math
from .duration import Multiplier
from .storage import FormatSpecification, StorageFormatManager


class NonreducedRatio(collections.abc.Sequence):
    """
    Nonreduced ratio.

    ..  container:: example

        Initializes from numbers:

        >>> abjad.NonreducedRatio((2, 4))
        NonreducedRatio((2, 4))

        >>> abjad.NonreducedRatio((2, 4, 2))
        NonreducedRatio((2, 4, 2))

    ..  container:: example

        Initializes from string:

        >>> abjad.NonreducedRatio("2:4")
        NonreducedRatio((2, 4))

        >>> abjad.NonreducedRatio("2:4:2")
        NonreducedRatio((2, 4, 2))

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_numbers",)

    ### INITIALIZER ###

    def __init__(self, numbers=(1, 1)):
        if isinstance(numbers, type(self)):
            numbers = numbers.numbers
        elif isinstance(numbers, str):
            strings = numbers.split(":")
            numbers = [int(_) for _ in strings]
        numbers = tuple(numbers)
        self._numbers = numbers

    ### SPECIAL METHODS ###

    def __contains__(self, argument):
        """
        Is true when ratio contains ``argument``.

        Returns true or false.
        """
        return argument in self._numbers

    def __eq__(self, argument):
        """
        Is true when ``argument`` is a nonreduced ratio with numerator and
        denominator equal to those of this nonreduced ratio.

        Returns true or false.
        """
        return StorageFormatManager.compare_objects(self, argument)

    def __getitem__(self, argument):
        """
        Gets item or slice identified by ``argument``.

        ..  container:: example

            >>> ratio = abjad.NonreducedRatio((2, 4, 2))
            >>> ratio[1]
            4

        Returns integer or tuple.
        """
        if isinstance(argument, slice):
            return tuple(self._numbers.__getitem__(argument))
        return self._numbers.__getitem__(argument)

    def __hash__(self):
        """
        Hashes non-reduced ratio.

        Required to be explicitly redefined on Python 3 if __eq__ changes.

        Returns integer.
        """
        return super().__hash__()

    def __iter__(self):
        """
        Iterates ratio.

        Returns generator.
        """
        return iter(self._numbers)

    def __len__(self):
        """
        Gets length of ratio.

        ..  container:: example

            >>> ratio = abjad.NonreducedRatio((2, 4, 2))
            >>> len(ratio)
            3

        Returns integer.
        """
        return len(self._numbers)

    def __repr__(self) -> str:
        """
        Gets interpreter representation.
        """
        return StorageFormatManager(self).get_repr_format()

    def __reversed__(self):
        """
        Iterates ratio in reverse.

        Returns generator.
        """
        return reversed(self._numbers)

    def __rtruediv__(self, number):
        """
        Divides ``number`` by ratio.

        ..  container:: example

            >>> 1 / abjad.Ratio((1, 1, 3))
            [Fraction(1, 5), Fraction(1, 5), Fraction(3, 5)]

        ..  container:: example

            >>> 1 / abjad.Ratio((1, 1, 3))
            [Fraction(1, 5), Fraction(1, 5), Fraction(3, 5)]

        ..  container:: example

            >>> 1.0 / abjad.Ratio((1, 1, 3))
            [0.2, 0.2, 0.6]

        Returns list of fractions or list of floats.
        """
        denominator = sum(self.numbers)
        factors = [quicktions.Fraction(_, denominator) for _ in self.numbers]
        result = [_ * number for _ in factors]
        return result

    __rdiv__ = __rtruediv__

    ### PRIVATE METHODS ###

    def _get_format_specification(self):
        return FormatSpecification(
            client=self,
            storage_format_args_values=[self.numbers],
            storage_format_is_indented=False,
            storage_format_keyword_names=[],
        )

    ### PUBLIC PROPERTIES ###

    @property
    def multipliers(self):
        """
        Gets multipliers of nonreduced ratio.

        ..  container:: example

            Nonreduced ratio of two numbers:

            >>> ratio = abjad.NonreducedRatio((2, 4))
            >>> ratio.multipliers
            (Multiplier(1, 3), Multiplier(2, 3))

        ..  container:: example

            Nonreduced ratio of three numbers:

            >>> ratio = abjad.NonreducedRatio((2, 4, 2))
            >>> ratio.multipliers
            (Multiplier(1, 4), Multiplier(1, 2), Multiplier(1, 4))

        Returns tuple of multipliers.
        """
        weight = sum(self.numbers)
        multipliers = [Multiplier((_, weight)) for _ in self.numbers]
        multipliers = tuple(multipliers)
        return multipliers

    @property
    def numbers(self):
        """
        Gets numbers of nonreduced ratio.

        ..  container:: example

            Nonreduced ratio of two numbers:

            >>> ratio = abjad.NonreducedRatio((2, 4))
            >>> ratio.numbers
            (2, 4)

        ..  container:: example

            Nonreduced ratio of three numbers:

            >>> ratio = abjad.NonreducedRatio((2, 4, 2))
            >>> ratio.numbers
            (2, 4, 2)

        Set to tuple of two or more numbers.

        Returns tuple of two or more numbers.
        """
        return self._numbers

    ### PUBLIC METHODS ###

    def count(self, argument):
        """
        Gets count of ``argument`` in ratio.

        Returns integer.
        """
        return self._numbers.count(argument)

    def index(self, argument):
        """
        Gets index of ``argument`` in ratio.

        Returns integer.
        """
        return self._numbers.index(argument)


class Ratio(NonreducedRatio):
    """
    Ratio.

    ..  container:: example

        Initializes from numbers:

        >>> abjad.Ratio((2, 4))
        Ratio((1, 2))

        >>> abjad.Ratio((2, 4, 2))
        Ratio((1, 2, 1))

    ..  container:: example

        Initializes from string:

        >>> abjad.Ratio("2:4")
        Ratio((1, 2))

        >>> abjad.Ratio("2:4:2")
        Ratio((1, 2, 1))

    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZER ###

    def __init__(self, numbers=(1, 1)):
        if isinstance(numbers, type(self)):
            numbers = numbers.numbers
        elif isinstance(numbers, str):
            strings = numbers.split(":")
            numbers = [int(_) for _ in strings]
        numbers = [int(_) for _ in numbers]
        gcd = math.greatest_common_divisor(*numbers)
        numbers = [_ // gcd for _ in numbers]
        self._numbers = tuple(numbers)

    ### SPECIAL METHODS ###

    def __eq__(self, argument):
        """
        Is true when ``argument`` equals ratio.

        ..  container:: example

            >>> ratio_1 = abjad.Ratio((1, 2, 1))
            >>> ratio_2 = abjad.Ratio((1, 2, 1))
            >>> ratio_3 = abjad.Ratio((2, 3, 3))

            >>> ratio_1 == ratio_1
            True

            >>> ratio_1 == ratio_2
            True

            >>> ratio_1 == ratio_3
            False

            >>> ratio_2 == ratio_1
            True

            >>> ratio_2 == ratio_2
            True

            >>> ratio_2 == ratio_3
            False

            >>> ratio_3 == ratio_1
            False

            >>> ratio_3 == ratio_2
            False

            >>> ratio_3 == ratio_3
            True

        """
        return super().__eq__(argument)

    def __getitem__(self, argument):
        """
        Gets item or slice identified by ``argument``.

        ..  container:: example

            >>> ratio = abjad.Ratio((2, 4, 2))
            >>> ratio[1]
            2

        Returns integer or tuple.
        """
        if isinstance(argument, slice):
            return tuple(self._numbers.__getitem__(argument))
        return self._numbers.__getitem__(argument)

    def __hash__(self):
        """
        Hashes ratio.

        Required to be explicitly redefined on Python 3 if __eq__ changes.

        Returns integer.
        """
        return super().__hash__()

    def __len__(self):
        """
        Gets length of ratio.

        ..  container:: example

            >>> ratio = abjad.Ratio((2, 4, 2))
            >>> len(ratio)
            3

        Returns integer.
        """
        return len(self._numbers)

    def __str__(self):
        """
        Gets string representation of ratio.

        ..  container:: example

            Ratio of two numbers:

            >>> str(abjad.Ratio((2, 4)))
            '1:2'

        ..  container:: example

            Ratio of three numbers:

            >>> str(abjad.Ratio((2, 4, 2)))
            '1:2:1'

        Returns string.
        """
        numbers = (str(x) for x in self.numbers)
        return ":".join(numbers)

    ### PUBLIC PROPERTIES ###

    @property
    def multipliers(self):
        """
        Gets multipliers of ratio.

        ..  container:: example

            Ratio of two numbers:

            >>> ratio = abjad.Ratio((2, 4))
            >>> ratio.multipliers
            (Multiplier(1, 3), Multiplier(2, 3))

        ..  container:: example

            Ratio of three numbers:

            >>> ratio = abjad.Ratio((2, 4, 2))
            >>> ratio.multipliers
            (Multiplier(1, 4), Multiplier(1, 2), Multiplier(1, 4))

        Returns tuple of multipliers.
        """
        weight = sum(self.numbers)
        multipliers = [Multiplier((_, weight)) for _ in self.numbers]
        multipliers = tuple(multipliers)
        return multipliers

    @property
    def numbers(self):
        """
        Gets numbers of ratio.

        ..  container:: example

            Ratio of two numbers:

            >>> ratio = abjad.Ratio((2, 4))
            >>> ratio.numbers
            (1, 2)

        ..  container:: example

            Ratio of three numbers:

            >>> ratio = abjad.Ratio((2, 4, 2))
            >>> ratio.numbers
            (1, 2, 1)

        Set to tuple of two or more numbers.

        Returns tuple of two or more numbers.
        """
        return self._numbers

    @property
    def reciprocal(self):
        """
        Gets reciprocal.

        ..  container:: example

            Gets reciprocal:

            >>> abjad.Ratio((3, 2)).reciprocal
            Ratio((2, 3))

            >>> abjad.Ratio((3, 2, 7)).reciprocal
            Ratio((7, 2, 3))

        Returns new ratio.
        """
        numbers = list(reversed(self.numbers))
        return type(self)(numbers)

    ### PUBLIC METHODS ###

    def partition_integer(self, n) -> typing.List[int]:
        """
        Partitions positive integer-equivalent ``n`` by ``ratio``.

        ..  container:: example

            >>> abjad.Ratio([1, 2]).partition_integer(10)
            [3, 7]

        ..  container:: example

            Partitions positive integer-equivalent ``n`` by ``ratio`` with negative
            parts:

            >>> abjad.Ratio([1, -2]).partition_integer(10)
            [3, -7]

        ..  container:: example

            Partitions negative integer-equivalent ``n`` by ``ratio``:

            >>> abjad.Ratio([1, 2]).partition_integer(-10)
            [-3, -7]

        ..  container:: example

            Partitions negative integer-equivalent ``n`` by ``ratio`` with negative
            parts:

            >>> abjad.Ratio([1, -2]).partition_integer(-10)
            [-3, 7]

        ..  container:: example

            More examples:

            >>> abjad.Ratio([1]).partition_integer(10)
            [10]

            >>> abjad.Ratio([1, 1]).partition_integer(10)
            [5, 5]

            >>> abjad.Ratio([1, -1, -1]).partition_integer(10)
            [3, -4, -3]

            >>> abjad.Ratio([1, 1, 1, 1]).partition_integer(-10)
            [-3, -2, -3, -2]

            >>> abjad.Ratio([1, 1, 1, 1, 1]).partition_integer(-10)
            [-2, -2, -2, -2, -2]

        Returns result with weight equal to absolute value of ``n``.
        """
        if not math.is_integer_equivalent_number(n):
            raise TypeError(f"is not integer-equivalent number: {n!r}.")
        ratio = self.numbers
        if not all(math.is_integer_equivalent_number(part) for part in ratio):
            message = f"some parts in {ratio!r} not integer-equivalent numbers."
            raise TypeError(message)
        result = [0]
        divisions = [float(abs(n)) * abs(part) / math.weight(ratio) for part in ratio]
        cumulative_divisions = math.cumulative_sums(divisions, start=None)
        for division in cumulative_divisions:
            rounded_division = int(round(division)) - sum(result)
            if division - round(division) == 0.5:
                rounded_division += 1
            result.append(rounded_division)
        result = result[1:]
        if math.sign(n) == -1:
            result = [-x for x in result]
        ratio_signs = [math.sign(x) for x in ratio]
        result = [pair[0] * pair[1] for pair in zip(ratio_signs, result)]
        return result
